# coding: utf-8
# Distributed under the terms of the MIT license.

""" This file implements the :class:`ComputeTask` class for handling
calculations on a single structure.

"""

import os
import sys
import shutil
import subprocess as sp
import glob
import time
import logging
import traceback as tb
from copy import deepcopy
from math import ceil
from psutil import virtual_memory

from matador import __version__
from matador.scrapers.castep_scrapers import cell2dict
from matador.scrapers.castep_scrapers import res2dict, castep2dict
from matador.calculators import CastepCalculator
from matador.crystal import Crystal
from matador.export import doc2cell, doc2param, doc2res
from matador.utils.errors import CriticalError, WalltimeError, InputError, CalculationError, MaxMemoryEstimateExceeded
from matador.utils.chem_utils import get_root_source
import matador.utils.print_utils

MATADOR_CUSTOM_TASKS = ['bulk_modulus', 'projected_bandstructure', 'pdispersion', 'all']

LOG = logging.getLogger('run3')
LOG.setLevel(logging.DEBUG)

__all__ = ["ComputeTask"]


class ComputeTask:
    """ The main use of this class is to call an executable on a given
    structure. The various parameters are passed to this class by the
    common entrypoints, run3 and ilustrado. It is unlikely that you
    will want to use this class directly. Each keyword is saved as
    an attribute for later use.

    Attributes:
        self.final_result (dict): stores the final result of the
            calculation, if it was successful.

    Note:
        By default, calculations are run inside a folder with the same
        name as the host (e.g. node12, or whatever). This decreases the
        load on parallel file systems such as Lustre.

    """

    def __init__(self, res, ncores, nnodes, node, **kwargs):
        """ Make the files to run the calculation and call the desired program.

        Parameters:
            res (str/dict): filename or input structure dict
            ncores (int): number of cores *per node* for mpirun call
            nnodes (int): number of nodes for mpirun call (if None, use 1)
            node (str): node name to run on (if None, run on localhost)

        Keyword arguments:
            param_dict (dict): dictionary of CASTEP parameters
            cell_dict (dict): dictionary of CASTEP cell options
            executable (str): name of binary to execute (DEFAULT: 'castep').
                Special string $seed will be parsed as the seedname,
                e.g. executable = 'pw6.x -i $seed.in > $seed.out' (requires mode='generic').
            mode (str): either 'castep' or 'generic' (DEFAULT: 'castep')
            custom_params (bool): use custom per-structure param file
                (DEFAULT: False)
            output_queue (multiprocessing.Queue): write results to queue rather than file.
            rough (int): number of small "rough" calculations (DEFAULT: 4)
            rough_iter (int): number of iterations per rough calculation
                (DEFAULT: 2)
            fine_iter (int): number of iterations per fine calculation
                (DEFAULT: 20)
            spin (bool): break spin symmetry in first calculation by amount specified
                (DEFAULT: None if not present, 5 if no argument)
            conv_cutoffs (:obj:`list` of :obj:`float`): list of cutoffs
                to use for SCF convergence test
            conv_kpts (:obj:`list` of :obj:`float`): list of kpt spacings
                to use for SCF convergence test
            kpts_1D (bool): treat z-direction as special and create
                kpt_grid [1 1 n_kz] (DEFAULT: False)
            noise (bool): add noise to the positions (DEFAULT: False)
            squeeze (bool/float): add an external pressure to the first steps (DEFAULT: False)
            archer (bool): force use of aprun over mpirun (DEFAULT: False)
            slurm (bool): force use of srun over mpirun (DEFAULT: False)
            intel (bool): force use of Intel mpirun-style calls (DEFAULT: False)
            redirect (str): file to redirect stdout to (DEFAULT: /dev/null unless debug).
            exec_test (bool): test executable with `<exec> --version`
                before progressing (DEFAULT: True)
            start (bool): begin calculation immediately or manually
                call it (DEFAULT: True)
            reopt (bool): whether to optimise one more time after
                success (DEFAULT: False)
            memcheck (bool): perform CASTEP dryrun to estimate memory
                usage, do not proceed if fails (DEFAULT: False)
            maxmem (int): maximum memory allowed in MB for memcheck
                (DEFAULT: None)
            killcheck (bool): check for file called $seed.kill during
                operation, and kill executable if present (DEFAULT: True)
            compute_dir (str): folder to run computations in; default is
                None (i.e. cwd), if not None, prepend paths with this folder
            verbosity (int): either 0, 1, 2 or >3, corresponding to ERROR, WARNING
                INFO and DEBUG logging levels.
            timings (`obj`:tuple: of `obj`:int:): tuple containing max and elapsed time in seconds

        Raises:
            WalltimeError: if desired/alotted walltime is exceeded, current run will be
                tidied up, ready to be restarted from intermediate state.
            CriticalError: if a fatal error occurs, failed run will be moved to bad_castep and
                no further calculations will be attempted.
            CalculationError: if a structure-level error occurs, causing the seed files to be moved
                to bad_castep.

        """
        # set defaults and update class with desired values
        prop_defaults = {'paths': None, 'param_dict': None, 'cell_dict': None, 'mode': 'castep', 'executable': 'castep',
                         'memcheck': False, 'rough': 4, 'rough_iter': 2, 'fine_iter': 20, 'spin': None, 'squeeze': False,
                         'output_queue': None, 'redirect': None, 'reopt': False, 'compute_dir': None, 'noise': False,
                         'custom_params': False, 'archer': False, 'maxmem': None, 'killcheck': True, 'kpts_1D': False,
                         'conv_cutoff': False, 'conv_kpt': False, 'slurm': False, 'intel': False,
                         'exec_test': True, 'timings': (None, None), 'start': True, 'verbosity': 1, 'polltime': 60,
                         'optados_executable': 'optados', 'run3_settings': dict(), 'workflow_kwargs': dict()}

        self.paths = None
        self.output_queue = None
        self.final_result = None
        self.executable = None

        self._first_run = True
        self._geom_max_iter_list = None
        self._squeeze_list = None
        self._max_iter = None
        self._num_rough_iter = None
        self._target_spacing = None
        self._target_pressure = None
        self._num_retries = 0
        self._max_num_retries = 2
        self.maxmem = None
        self.cell_dict = None
        self.param_dict = None
        self.res_dict = None
        self.calc_doc = None

        # save all keyword arguments as attributes
        self.__dict__.update(prop_defaults)
        for arg in kwargs:
            if arg in prop_defaults:
                self.__dict__.update({arg: kwargs[arg]})

        if self.maxmem is None and self.memcheck:
            self.maxmem = float(virtual_memory().available) / 1024**2

        # scrape and save seed name
        if isinstance(res, str):
            self.seed = res
            self.seed = self.seed.split('/')[-1].replace('.res', '')
        else:
            self.seed = get_root_source(res)
        self._input_res = res

        # set up logging; by default, write DEBUG level into a file called logs/$seed.log
        # and write desired verbosity to stdout (default WARN).
        if self.verbosity == 0:
            loglevel = logging.ERROR
        elif self.verbosity == 1:
            loglevel = logging.WARN
        elif self.verbosity == 2:
            loglevel = logging.INFO
        else:
            loglevel = logging.DEBUG

        if not os.path.isdir('logs'):
            os.mkdir('logs')

        LOG.handlers = []

        if self.verbosity > 1:
            stdout_handler = logging.StreamHandler(sys.stdout)
            stdout_handler.setLevel(loglevel)
            stdout_handler.setFormatter(logging.Formatter('%(asctime)s - %(name)s | %(levelname)8s: %(message)s'))
            LOG.addHandler(stdout_handler)

        logname = os.path.abspath('logs/{}.log'.format(self.seed))
        file_handler = logging.FileHandler(logname, mode='a')
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(logging.Formatter('%(asctime)s - %(name)s | %(levelname)8s: %(message)s'))
        LOG.addHandler(file_handler)

        splash_screen = r"""
                           _____
          _ __ _   _ _ __ |___ /
         | '__| | | | '_ \  |_ \
         | |  | |_| | | | |___) |
         |_|   \__,_|_| |_|____/

        """

        LOG.info(splash_screen)
        LOG.info('matador version: {}'.format(__version__))
        LOG.info('Host info: {}.'.format('-'.join(os.uname())))
        LOG.info('Run started for seed {}'.format(self.seed))

        # set up compute parameters
        self.ncores = ncores
        self.nnodes = nnodes
        self.node = node
        self.conv_cutoff_bool = isinstance(self.conv_cutoff, list)
        self.conv_kpt_bool = isinstance(self.conv_kpt, list)
        self.enough_memory = True
        self.success = None
        self._mpi_library = None
        self.root_folder = os.getcwd()

        if self.timings is not None:
            self.max_walltime = self.timings[0]
            self.start_time = self.timings[1]

        if self.max_walltime is not None:
            LOG.debug('Starting at {start}, max walltime allowed is {walltime}'.format(
                start=self.start_time, walltime=self.max_walltime))
            LOG.debug('{} s remaining.'.format(self.max_walltime - (time.time() - self.start_time)))

        self._redirect_filename = None

        if self.paths is None:
            self.paths = {}
            self.paths['completed_dir'] = 'completed'
            self.paths['failed_dir'] = 'bad_castep'
            self.paths['memory_fname'] = 'memory_exceeded.txt'
            self.paths['jobs_fname'] = 'jobs.txt'
            self.paths['completed_fname'] = 'finished_cleanly.txt'
            self.paths['failures_fname'] = 'failures.txt'

        elif 'completed_dir' not in self.paths:
            raise RuntimeError('Invalid paths: {}'.format(self.paths))

        self._set_input_structure(res)

        if self.start:
            self.begin()
        else:
            LOG.info('Waiting for `begin()` method to be called...')

    def _set_input_structure(self, res):
        """ Set the input structure to the given dictionary, or read it from
        the given file.

        Parameters:
            res (str/dict): either the scraped structure or the filename of a res file.

        """

        # read in initial structure and skip if failed
        if isinstance(res, str):
            self.res_dict, success = res2dict(res, db=False, verbosity=self.verbosity)
            if not success:
                msg = 'Unable to parse initial res file, error: {res_dict}'.format(res_dict=self.res_dict)
                raise CalculationError(msg)

        elif isinstance(res, dict):
            # check res is a valid crystal
            try:
                Crystal(res)
            except Exception as exc:
                msg = 'Unable to convert res to Crystal: {}'.format(exc)
                LOG.error(msg)
                LOG.error('Traceback:\n{}'.format(tb.format_exc()))
                raise CalculationError(msg)

            self.res_dict = res

    def begin(self):
        """ Run the prepared ComputeTask. Catches CalculationError objects
        and cleans up, passing all other errors upwards.

        """

        try:
            try:
                # run through CASTEP specific features
                if self.mode == 'castep':
                    calculation_parameters = {**self.cell_dict, **self.param_dict}
                    self.calculator = CastepCalculator(calculation_parameters)
                    self.success = self.run_castep()
                # otherwise run generic script
                else:
                    self.success = self.run_generic()

            except CalculationError as exc:
                self.success = False
                self._finalise_result()
                raise exc

        except Exception as exc:
            LOG.error('Process raised {} with message {}.'.format(type(exc), exc))
            if isinstance(exc, CalculationError):
                LOG.error('Full traceback:\n{}'.format(tb.format_exc()))
            for handler in LOG.handlers[:]:
                handler.close()
            raise exc

        if self.success:
            LOG.info('ComputeTask finished successfully for {seed}'.format(seed=self.seed))
        else:
            LOG.info('ComputeTask failed cleanly for {seed}'.format(seed=self.seed))
        for handler in LOG.handlers[:]:
            handler.close()

    def run_castep(self):
        """ Set up and run CASTEP calculation on the prepared structure,
        `self.res_dict`, using the parameters in `self.cell_dict` and
        `self.param_dict`.

        Raises:
            WalltimeError: if max_walltime is exceeded.
            CriticalError: if no further calculations should be performed
                on this thread.
            CalculationError: if this structure errored in some way, but
                others will hopefully be okay.

        Returns:
            bool: True if calculations were successful, False otherwise.
                In the case of convergence tests, this is always True
                unless every calculation fails.

        """
        success = False

        if self.exec_test:
            self.test_exec()

        # pre-processing
        if self.kpts_1D:
            LOG.debug('1D kpoint grid requested.')
            if 'kpoints_mp_spacing' not in self.cell_dict:
                raise CriticalError('kpoints_mp_spacing not found, but kpts_1D requested...')
            self._target_spacing = deepcopy(self.cell_dict['kpoints_mp_spacing'])

        if self.squeeze:
            self._target_pressure = deepcopy(self.cell_dict['external_pressure'])

        if self.noise:
            from matador.utils.cell_utils import add_noise
            LOG.info('Adding noise to the positions in the cell')
            self.res_dict = add_noise(self.res_dict, amplitude=0.1)

        self.calc_doc = deepcopy(self.res_dict)

        # update global doc with cell and param dicts for folder, and verify
        bad_keys = ['atom_types', 'positions_frac', 'positions_abs', 'lattice_cart', 'lattice_abc']
        self.cell_dict = {key: self.cell_dict[key] for key in self.cell_dict if key not in bad_keys}
        if "write_cell_structure" not in self.cell_dict:
            self.cell_dict["write_cell_structure"] = True

        self.calc_doc.update(self.cell_dict)
        self.calc_doc.update(self.param_dict)
        self.calculator.verify_calculation_parameters(self.calc_doc, self.res_dict)

        try:
            LOG.info('Struture: {}'.format(Crystal(self.calc_doc)))
        except Exception as exc:
            LOG.warning('Unable to convert structure to Crystal... {}'.format(exc))

        # now verify the structure itself
        self.calculator.verify_simulation_cell(self.res_dict)

        # do memcheck, if desired, and only continue if enough memory is free
        if self.memcheck:
            memory_usage_estimate = self.do_memcheck(self.calc_doc, self.seed)
            mem_string = ('Memory estimate / Available memory (MB): {:8.0f} / {:8.0f}'
                          .format(memory_usage_estimate, self.maxmem))
            LOG.info(mem_string)

            if memory_usage_estimate > 0.9 * self.maxmem:
                msg = 'Structure {} failed memcheck, skipping... '.format(self.seed) + mem_string
                with open(self.paths['memory_fname'], 'a') as f:
                    f.write("{} {:.2f}GB/{:.2f}GB\n".format(self.seed, memory_usage_estimate / 1024, self.maxmem / 1024))

                raise MaxMemoryEstimateExceeded(msg)

        try:
            # run convergence tests
            if any([self.conv_cutoff_bool, self.conv_kpt_bool]):
                success = self.run_convergence_tests(self.calc_doc)

            # perform relaxation
            elif self.calc_doc['task'].upper() in ['GEOMETRYOPTIMISATION', 'GEOMETRYOPTIMIZATION']:
                success = self.run_castep_relaxation()

            elif self.calc_doc['task'].upper() in ['PHONON', 'THERMODYNAMICS']:
                from matador.workflows.castep import castep_full_phonon
                success = castep_full_phonon(self, self.calc_doc, self.seed, **self.workflow_kwargs)

            elif self.calc_doc['task'].upper() in ['SPECTRAL']:
                from matador.workflows.castep import castep_full_spectral
                success = castep_full_spectral(self, self.calc_doc, self.seed, **self.workflow_kwargs)

            elif self.calc_doc['task'].upper() in ['BULK_MODULUS']:
                from matador.workflows.castep import castep_elastic
                success = castep_elastic(self, self.calc_doc, self.seed, **self.workflow_kwargs)

            # run in singleshot mode, i.e. just call CASTEP on the seeds
            else:
                success = self.run_castep_singleshot(self.calc_doc, self.seed, keep=True)

            self._first_run = False

            return success

        finally:
            # always cd back to root folder, in case we have ended up on the wrong place
            if self.compute_dir is not None:
                os.chdir(self.root_folder)
                self.remove_compute_dir_if_finished(self.compute_dir)

    def run_generic(self, intermediate=False, mv_bad_on_failure=True):
        """ Run a generic mpi program on the given seed. Files from
        completed runs are moved to "completed" (unless intermediate
        is True) and failed runs to "bad_castep".

        Keyword arguments:
            intermediate (bool): whether we want to run more calculations
                on the output of this, i.e. whether to move to completed
                or not.
            mv_bad_on_failure (bool): whether to move files to bad_castep on
                failure, or leave them in place.

        Returns:
            bool: True if calculations progressed without error.

        """
        LOG.info('Calling executable {exe} MPI program on {seed}'.format(exe=self.executable, seed=self.seed))
        try:
            seed = self._input_res
            if '.' in seed:
                seed = seed.split('.')[-2]
                input_ext = seed.split('.')[-1]
            else:
                input_ext = ''

            assert isinstance(seed, str)
            self.cp_to_input(seed, ext=input_ext, glob_files=True)

            self._setup_compute_dir(self.seed, self.compute_dir, generic=True)
            if self.compute_dir is not None:
                os.chdir(self.compute_dir)

            _process = self.run_command(seed)
            self._handle_process(_process, check_walltime=True)

            if not intermediate:
                LOG.info('Writing results of generic call to res file and tidying up.')
                self.mv_to_completed(seed, keep=True, completed_dir=self.paths['completed_dir'])

            LOG.info('Executable {exe} finished cleanly.'.format(exe=self.executable))
            self._first_run = False

            return True

        except Exception as err:
            self._first_run = False
            LOG.error('Caught error inside run_generic: {error}.'.format(error=err))
            if mv_bad_on_failure:
                self.mv_to_bad(seed)
            raise err

        finally:
            if self.compute_dir is not None:
                os.chdir(self.root_folder)

    def run_castep_relaxation(self, intermediate=False):
        """ Set up a structural relaxation that is restarted intermittently
        in order to re-mesh the kpoint grid. Completed calculations are moved
        to the "completed" folder, and failures to "bad_castep".

        Keyword arguments:
            intermediate (bool): whether we want to run more calculations
                on the output of this, i.e. whether to move to completed
                or not.

        Returns:
            bool: True iff structure was optimised, False otherwise.

        Raises:
            CalculationError: if structure-level error occured.
            CriticalError: if fatal global error occured.
            WalltimeError: if walltime was reached, and jobs need to stop.

        """
        try:
            return self._relax(intermediate=intermediate)

        except WalltimeError as err:
            raise err

        except Exception as err:
            self._finalise_result(intermediate)
            raise err

        finally:
            if self.compute_dir is not None:
                os.chdir(self.root_folder)

    def _relax(self, intermediate=False):
        """ Set up a structural relaxation that is restarted intermittently
        in order to re-mesh the kpoint grid. Completed calculations are moved
        to the "completed" folder, and failures to "bad_castep".

        Keyword arguments:
            intermediate (bool): whether we want to run more calculations
                on the output of this, i.e. whether to move to completed
                or not.

        Returns:
            bool: True iff structure was optimised, False otherwise.

        Raises:
            CalculationError: if structure-level error occured.
            CriticalError: if fatal global error occured.
            WalltimeError: if walltime was reached, and jobs need to stop.

        """
        self._setup_relaxation()
        seed = self.seed
        rerun = False
        # iterate over geom iter blocks
        for ind, num_iter in enumerate(self._geom_max_iter_list):

            # preprocess last step that did not finish geom opt
            if self.reopt and rerun:
                # if we're now reoptimising after a success in last step
                # use the fine iter value
                num_iter = self.fine_iter
                LOG.info('Last step was successful, performing one last relaxation...')

            # update the geom_max_iter to use with either the number in iter_list, or the overriden value
            self.calc_doc['geom_max_iter'] = num_iter

            # delete any existing files and write new ones
            if self.squeeze:
                squeeze = int(self._squeeze_list[ind]) * float(self.squeeze)
            else:
                squeeze = None
            self._update_input_files(self.seed, self.calc_doc, squeeze=squeeze)

            if self.max_walltime is not None and self.start_time is None:
                msg = 'Somehow initial start time was not found'
                LOG.critical(msg)
                raise CriticalError(msg)

            # run CASTEP
            _process = self.run_command(seed)

            # will throw errors if the process fails
            output_filename = "{}.{}".format(seed, 'castep')

            self._handle_process(
                _process, expected_fname=output_filename, check_walltime=True
            )

            # check for errors and try to correct for them
            errors_present, errors, remedy = self._catch_castep_errors(_process)
            skip_postprocess = remedy is not None

            # try to read the CASTEP file
            opti_dict, success = castep2dict(output_filename, db=False, verbosity=self.verbosity)

            if errors_present:
                msg = 'Failed to optimise {} as CASTEP crashed with error:'.format(seed)
                msg += errors
                LOG.warning(msg)
                if isinstance(opti_dict, dict):
                    self._update_castep_output_files(seed, opti_dict)
                if remedy is not None and self._num_retries <= self._max_num_retries:
                    LOG.warning('Attempting to recover using {}'.format(remedy))
                else:
                    raise CalculationError(msg)

            if not success and remedy is None and isinstance(opti_dict, Exception):
                msg = 'Failed to parse CASTEP file... {}'.format(opti_dict)
                LOG.warning(msg)
                raise CalculationError(msg)
            if isinstance(opti_dict, Exception) and remedy is not None:
                opti_dict = {'optimised': False}

            LOG.info('Intermediate calculation completed successfully: total relaxation steps now = {} '
                     .format(opti_dict.get('geom_iter')))

            # scrub keys that need to be rescraped
            keys_to_remove = ['kpoints_mp_spacing', 'kpoints_mp_grid', 'species_pot',
                              'sedc_apply', 'sedc_scheme', 'cell_constraints']
            for key in keys_to_remove:
                if key in opti_dict:
                    del opti_dict[key]

            if not skip_postprocess:
                # now post-process the last step
                # did we try to rerun, and are now no longer optimised?
                # then reset, and go again...
                if self.reopt and rerun and not opti_dict['optimised']:
                    rerun = False
                    self._update_castep_output_files(seed, opti_dict)

                # are we optimised, but haven't yet rerun?
                # then set prepare to do one more full relaxation
                if self.reopt and not rerun and opti_dict['optimised']:
                    rerun = True

                    # disable squeezing if we've already reached optimisation
                    if self.squeeze:
                        for jnd in range(ind, len(self._squeeze_list)):
                            self._squeeze_list[jnd] = False
                    self._update_castep_output_files(seed, opti_dict)

                # or did the relaxation complete successfuly, including rerun?
                elif (not self.reopt or rerun) and opti_dict['optimised']:
                    LOG.info('Successfully relaxed {}'.format(seed))
                    self._update_castep_output_files(seed, opti_dict)
                    break

                # reached maximum number of steps
                elif ind == len(self._geom_max_iter_list) - 1:
                    msg = 'Failed to optimise {} after {} steps'.format(seed, sum(self._geom_max_iter_list))
                    LOG.info(msg)
                    raise CalculationError(msg)

                # if we weren't successful, then preprocess the next step

                # set atomic_init_spins with value from CASTEP file, if it exists
                if 'mulliken_spins' in opti_dict:
                    self.calc_doc['atomic_init_spins'] = opti_dict['mulliken_spins']

                # CASTEP prints cell constraints in castep file when running with symmetry
                # its useful to disable these and use either the initial constraints, or none
                if 'cell_constraints' in self.cell_dict:
                    self.calc_doc['cell_constraints'] = self.cell_dict['cell_constraints']

                # if writing out cell, use it for higher precision positions and lattice_cart
                if self.calc_doc.get('write_cell_structure') and os.path.isfile('{}-out.cell'.format(seed)):
                    cell_dict, success = cell2dict('{}-out.cell'.format(seed),
                                                   verbosity=self.verbosity,
                                                   db=False, positions=True, lattice=True)
                    if success:
                        opti_dict['lattice_cart'] = list(cell_dict['lattice_cart'])
                        if 'positions_frac' in cell_dict and 'atom_types' in cell_dict:
                            opti_dict['positions_frac'] = list(cell_dict['positions_frac'])
                            opti_dict['atom_types'] = list(cell_dict['atom_types'])

                LOG.info('N = {iters:03d} | |F| = {d[max_force_on_atom]:5.5f} eV/A | '
                         'S = {pressure:5.5f} GPa | H = {enthalpy_per_atom:5.5f} eV/atom'
                         .format(d=opti_dict,
                                 pressure=opti_dict.get('pressure', 0.0),
                                 enthalpy_per_atom=opti_dict.get('enthalpy_per_atom', 0.0),
                                 iters=opti_dict.get('geom_iter', 0)))

            # if there were errors that can be remedied, now is the time to do it
            # this will normally involve changing a parameter to avoid future failures
            self.calc_doc.update(opti_dict)

            if remedy is not None:
                LOG.info('Trying to remedy error with {}'.format(remedy))
                remedy(self.calc_doc)
                self._num_retries += 1

        return self._finalise_result(intermediate=intermediate)

    def run_castep_singleshot(self, calc_doc, seed, keep=True, intermediate=False):
        """ Perform a singleshot calculation with CASTEP. Singleshot runs do not
        attempt to remedy any errors raised.

        Files from completed runs are moved to `completed`, if not
        in intermediate mode, and failed runs to `bad_castep`.

        Parameters:
            calc_doc (dict): dictionary containing parameters and structure
            seed (str): structure filename

        Keyword arguments:
            intermediate (bool): whether we want to run more calculations
                on the output of this, i.e. whether to move to completed
                or not.
            keep (bool): whether to keep intermediate files e.g. .bands

        Returns:
            bool: True iff SCF completed successfully, False otherwise.

        """
        LOG.info('Performing single-shot CASTEP run on {}, with task: {}'.format(seed, calc_doc['task']))
        try:
            self._singleshot(calc_doc, seed, keep=keep, intermediate=intermediate)

        except WalltimeError as err:
            raise err

        except Exception as err:
            self.mv_to_bad(seed)
            if not keep:
                self.tidy_up(seed)
            raise err

        finally:
            if self.compute_dir is not None:
                os.chdir(self.root_folder)

    def _singleshot(self, calc_doc, seed, keep=True, intermediate=False):
        """ Perform a singleshot calculation with CASTEP. Singleshot runs do not
        attempt to remedy any errors raised.

        Files from completed runs are moved to `completed`, if not
        in intermediate mode, and failed runs to `bad_castep`.

        Parameters:
            calc_doc (dict): dictionary containing parameters and structure
            seed (str): structure filename

        Keyword arguments:
            intermediate (bool): whether we want to run more calculations
                on the output of this, i.e. whether to move to completed
                or not.
            keep (bool): whether to keep intermediate files e.g. .bands

        Returns:
            bool: True iff SCF completed successfully, False otherwise.

        """

        self.cp_to_input(seed)
        self._setup_compute_dir(seed, self.compute_dir, custom_params=self.custom_params)
        if self.compute_dir is not None:
            os.chdir(self.compute_dir)

        self._update_input_files(seed, calc_doc)
        doc2res(calc_doc, self.seed, info=False, hash_dupe=False, overwrite=True)

        # run CASTEP
        _process = self.run_command(seed)
        output_filename = "{}.{}".format(seed, 'castep')
        self._handle_process(
            _process, expected_fname=output_filename, check_walltime=True
        )

        # check for errors and report them
        errors_present, errors, _ = self._catch_castep_errors(_process)
        if errors_present:
            raise CalculationError(
                'CASTEP run on {} failed with errors: {}'.format(seed, errors)
            )

        # scrape dict but ignore the results
        results_dict, success = castep2dict(output_filename, db=False, verbosity=self.verbosity)
        if not success:
            raise CalculationError(
                'Error scraping CASTEP file {}: {}'.format(seed, results_dict)
            )

        self._update_castep_output_files(seed)

        if not intermediate:
            LOG.info('Writing results of singleshot CASTEP run to res file and tidying up.')
            doc2res(results_dict, seed, hash_dupe=False, overwrite=True)
            self.mv_to_completed(seed, keep=keep, completed_dir=self.paths['completed_dir'])
            self.tidy_up(seed)

        return success

    @staticmethod
    def validate_calc_doc(calc_doc, required, forbidden):
        """ Remove keys inside forbidden from calc_doc, and error
        if a required key is missing.

        Parameters:
            calc_doc (dict): dictionary of structure and parameters.
            required (list): list of required key strings.
            forbidden (list): list of forbidden keys.

        Raises:
            AssertionError: if required key is missing.

        """
        for keyword in forbidden:
            if keyword in calc_doc:
                del calc_doc[keyword]

        failures = []
        for keyword in required:
            if keyword not in calc_doc:
                failures.append(keyword)
        if failures:
            raise InputError('The following keywords are required for workflow: {}'.format(', '.join(failures)))

    @staticmethod
    def get_seekpath_compliant_input(calc_doc, spacing, debug=False):
        """ Return seekpath cell/kpoint path for the given cell and spacing.

        Parameters:
            calc_doc (dict): structural and calculation parameters.
            spacing (float): desired kpoint path spacing.

        Returns:
            (dict, list): dictionary containing the standardised unit cell
                and list containing the kpoints.

        """
        LOG.info('Getting seekpath cell and kpoint path.')
        from matador.utils.cell_utils import get_seekpath_kpoint_path
        LOG.debug('Old lattice: {}'.format(Crystal(calc_doc)))
        prim_doc, kpt_path, _ = get_seekpath_kpoint_path(calc_doc,
                                                         spacing=spacing,
                                                         symmetry_tol=calc_doc.get('symmetry_tol'),
                                                         debug=debug)
        LOG.debug('New lattice: {}'.format(Crystal(calc_doc)))
        return prim_doc, kpt_path

    def run_convergence_tests(self, calc_doc):
        """ Run kpoint and cutoff_energy convergence tests based on
        options passed to ComputeTask.

        Parameters:
            calc_doc (dict): the structure to converge.

        Returns:
            bool: True unless every single calculation failed.

        """
        LOG.info('Performing convergence tests...')
        from matador.utils.cell_utils import get_best_mp_offset_for_cell
        successes = []
        cached_cutoff = calc_doc['cut_off_energy']
        if self.conv_cutoff_bool:
            # run series of singlepoints for various cutoffs
            LOG.info('Running cutoff convergence...')
            for cutoff in self.conv_cutoff:
                LOG.info('{} eV... '.format(cutoff))
                calc_doc.update({'cut_off_energy': cutoff})
                self.paths['completed_dir'] = 'completed_cutoff'
                seed = self.seed + '_' + str(cutoff) + 'eV'
                success = self.run_castep_singleshot(calc_doc, seed, keep=False)
                successes.append(success)
        if self.conv_kpt_bool:
            # run series of singlepoints for various cutoffs
            LOG.info('Running kpt convergence tests...')
            calc_doc['cut_off_energy'] = cached_cutoff
            for kpt in self.conv_kpt:
                LOG.info('{} 1/A... '.format(kpt))
                calc_doc.update({'kpoints_mp_spacing': kpt})
                calc_doc['kpoints_mp_offset'] = get_best_mp_offset_for_cell(calc_doc)
                LOG.debug('Using offset {}'.format(calc_doc['kpoints_mp_offset']))
                self.paths['completed_dir'] = 'completed_kpts'
                seed = self.seed + '_' + str(kpt) + 'A'
                success = self.run_castep_singleshot(calc_doc, seed, keep=False)
                successes.append(success)
        return any(successes)

    def parse_executable(self, seed):
        """ Turn executable string into list with arguments to be executed.

        Example:
            With `self.executable='castep17'` and `seed='test'`, `['castep17', 'test']`
            will be returned.

        Example:
            With `self.executable='pw6.x -i $seed.in > $seed.out'` and `seed='test'`,
            `['pw6.x', '-i', 'test.in', '>' 'test.out']` will be returned.

        Parameters:
            seed (str): filename to replace $seed with in command.

        Returns:
            :obj:`list` of :obj:`str`: list called by subprocess.POpen.

        """
        if isinstance(self.executable, str):
            executable = self.executable.split()
        command = []
        found_seed = False
        for item in executable:
            if '$seed' in item:
                item = item.replace('$seed', seed)
                found_seed = True
            command.append(item)
        if not found_seed:
            command.append(seed)

        if self.redirect is not None:
            self._redirect_filename = self.redirect.replace('$seed', seed)
        else:
            self._redirect_filename = None

        LOG.debug('Executable string parsed as {}'.format(command))

        return command

    def test_exec(self):
        """ Test if <executable> --version returns a valid string.

        Raises:
            CriticalError: if executable not found.

        """
        LOG.info('Testing executable "{executable}".'.format(executable=self.executable))

        try:
            proc = self.run_command('--version')

        except FileNotFoundError:
            LOG.critical('Unable to call mpirun/aprun/srun, currently selected: {}'.format(self.mpi_library))
            message = 'Please check initialistion of ComputeTask object/CLI args.'
            LOG.debug('Raising CriticalError with message: {message}'.format(message=message))
            raise CriticalError(message)

        out, errs = proc.communicate()
        out = out.decode('utf-8')
        errs = errs.decode('utf-8')
        if 'CASTEP version' not in out:
            # this is an OpenMPI error that occurs when hyperthreading is enabled
            # best way to handle is to half the number of procs available
            if 'not enough slots' in errs:
                err_string = ('MPI library tried to use too many cores and failed, '
                              'rescaling core count and re-running with {} cores...'.format(int(self.ncores/2)))
                LOG.warning(err_string)
                LOG.warning('stdout: {stdout}'.format(stdout=out))
                LOG.warning('sterr: {stderr}'.format(stderr=errs))
                if self.ncores >= 2:
                    self.ncores = int(self.ncores/2)
                else:
                    raise CriticalError(err_string)
                self.test_exec()

            else:
                err_string = 'Executable `{}` failed testing: does it support --version flag?'.format(self.executable)
                LOG.critical(err_string)
                LOG.critical('stdout: {stdout}'.format(stdout=out))
                LOG.critical('sterr: {stderr}'.format(stderr=errs))
                LOG.debug('Raising CriticalError.')
                raise CriticalError(err_string)

        if errs:
            LOG.info('Executable {} passed test, but stderr contains the following:'.format(self.executable))
            LOG.info(errs)

        try:
            version_string = out.split('\n')
            for line in version_string:
                if 'CASTEP version' in line:
                    version = line.strip().split(' ')[-1]
                    version = float(version)

            LOG.info('CASTEP version: {}'.format(version))
            if version < 18:
                LOG.warning('Using CASTEP version {}: '
                            'some features may not work, consider updating to at least CASTEP 18'.format(version))
        except Exception:
            LOG.warning('Unable to detect CASTEP version: '
                        'some features may not work unless you are using at least version 18.')

    @property
    def mpi_library(self):
        """ Property to store/compute desired MPI library. """
        if self._mpi_library is None:
            self._mpi_library = self.set_mpi_library()
        return self._mpi_library

    def set_mpi_library(self):
        """ Combines command-line MPI arguments into string and calls
        MPI library detection is no args are present.
        """
        guessed_version = self.detect_mpi()
        LOG.info('Detected {} MPI.'.format(guessed_version))
        if sum([self.archer, self.intel, self.slurm]) > 1:
            message = 'Conflicting command-line arguments for MPI library have been supplied, exiting.'
            LOG.critical(message)
            raise CriticalError(message)

        if self.archer:
            if guessed_version != 'archer':
                message = ('Detected {} MPI, but user asked to use aprun... please check your environment.'
                           .format(self._mpi_library))
                LOG.critical(message)
                raise CriticalError(message)
            return 'archer'

        if self.intel:
            if guessed_version != 'intel':
                message = ('Detected {} MPI, but user asked to use Intel MPI... please check your environment.'
                           .format(self._mpi_library))
                LOG.critical(message)
                raise CriticalError(message)
            return 'intel'

        if self.slurm:
            if guessed_version != 'slurm':
                message = 'Detected {} MPI, but user asked to use srun... continuing with srun.'.format(self._mpi_library)
                LOG.warning(message)
            return 'slurm'

        return guessed_version

    @staticmethod
    def detect_mpi():
        """ Test which mpi library is being used when `mpirun`.

        Returns:
            mpi_library (str): 'intel', 'archer', or 'default'.

        """
        # check first for existence of mpirun command, then aprun if that fails
        try:
            try:
                mpi_version_string = str(sp.check_output('mpirun --version', shell=True))
            except sp.CalledProcessError:
                LOG.info('Failed to find mpirun, checking aprun...')
                mpi_version_string = str(sp.check_output('aprun --version', shell=True))
        except Exception as exc:
            msg = 'Failed to find mpirun or aprun.'
            LOG.critical(msg)
            LOG.debug('Error message: {exc}'.format(exc=exc))
            raise CriticalError(msg)
        if 'Intel' in mpi_version_string:
            mpi_version = 'intel'
        elif 'aprun' in mpi_version_string:
            mpi_version = 'archer'
        elif 'Open MPI' in mpi_version_string:
            mpi_version = 'default'
        else:
            LOG.debug('Could not detect MPI library so using default (OpenMPI), version string was: {response}'
                      .format(response=mpi_version_string))
            mpi_version = 'default'

        LOG.info('Using {version} MPI library.'.format(version=mpi_version))
        return mpi_version

    def do_memcheck(self, calc_doc, seed):
        """ Perform a CASTEP dryrun to estimate memory usage.

        Parameters:
            calc_doc (dict): dictionary of structure and CASTEP parameters
            seed (str): filename for structure

        Returns:
            bool: True if the memory estimate is <90% of node RAM or
                `self.maxmem`, if set

        """
        LOG.info('Performing memory check for {seed}'.format(seed=seed))
        memcheck_seed = seed + '_memcheck'

        memcheck_doc = deepcopy(calc_doc)

        if memcheck_doc['task'] in MATADOR_CUSTOM_TASKS:
            memcheck_doc['task'] = 'singlepoint'

        doc2param(memcheck_doc, memcheck_seed, hash_dupe=False)
        doc2cell(memcheck_doc, memcheck_seed, hash_dupe=False)

        LOG.debug('Running CASTEP dryrun.')
        self.executable += ' --dryrun'
        process = self.run_command(memcheck_seed)
        process.communicate()
        self.executable = self.executable.replace(' --dryrun', '')

        results, success = castep2dict(memcheck_seed + '.castep', db=False, verbosity=self.verbosity)

        for _file in glob.glob(memcheck_seed + '*'):
            if _file.endswith('.res'):
                continue
            else:
                os.remove(_file)

        if not success or 'estimated_mem_MB' not in results:
            msg = 'CASTEP dryrun failed with output {results}'.format(results=results)
            raise MaxMemoryEstimateExceeded(msg)

        estimate = results['estimated_mem_per_process_MB']
        if self.ncores is not None:
            estimate *= self.ncores
        if self.nnodes is not None:
            estimate *= self.nnodes

        return estimate

    def run_command(self, seed):
        """ Calls executable on seed with desired number of cores.

        Parameters:
            seed (str): seedname to pass append to CASTEP command,
                e.g. <seed> or --version.

        Returns:
            subprocess.Popen: process to run.

        """
        command = self.parse_executable(seed)
        if self.nnodes is None or self.nnodes == 1:
            if self.node is None:
                if self.ncores == 1 and self.node is None:
                    command = ['nice', '-n', '15'] + command
                elif self.mpi_library == 'archer':
                    command = ['aprun', '-n', str(self.ncores)] + command
                elif self.mpi_library == 'slurm':
                    command = ['srun'] + command
                elif self.mpi_library in ['intel', 'default']:
                    command = ['mpirun', '-n', str(self.ncores)] + command
                else:
                    command = ['mpirun', '-n', str(self.ncores)] + command
            elif self.node is not None:
                cwd = os.getcwd()
                command = ['ssh', '{}'.format(self.node), 'cd', '{};'.format(cwd), 'mpirun', '-n',
                           str(self.ncores)] + command
        else:
            if self.mpi_library == 'archer':
                command = [
                    'aprun', '-n',
                    str(self.ncores * self.nnodes), '-N',
                    str(self.ncores), '-S', '12', '-d', '1'
                ] + command
            elif self.mpi_library == 'slurm':
                command = ['srun'] + command
            elif self.mpi_library == 'intel':
                command = ['mpirun', '-n', str(self.ncores * self.nnodes), '-ppn', str(self.ncores)] + command
            else:
                command = ['mpirun', '-n', str(self.ncores * self.nnodes), '-npernode', str(self.ncores)] + command

        if self.ncores > 1 and 'run' not in ' '.join(command):
            raise RuntimeError('Issue running command {} in parallel, this is probably a matador bug!'.format(command))

        stdout = sp.PIPE
        stderr = sp.PIPE

        if self._redirect_filename is not None:
            LOG.info('Redirecting output to {redirect}'.format(redirect=self._redirect_filename))
            redirect_file = open(self._redirect_filename, 'w')
            stdout = redirect_file

        LOG.info('Running {}'.format(command))
        process = sp.Popen(command, shell=False, stdout=stdout, stderr=stderr)
        try:
            redirect_file.close()
        except Exception:
            pass

        return process

    def _handle_process(self, process, check_walltime=False, expected_fname=None):
        """ Poll, wait, communicate with running process, with optional
        checks on walltime and file creation.

        Parameters:
            process (subprocess.Popen): process object to handle.

        Keyword arguments:
            check_walltime (bool): timeout the process 2 x :attr:`polltime` before
                the allotted walltime, and clean up any files.
            expected_fname (str): the filename to check for, assuming filename

        Raises:
            CalculationError: if calculation reports its own failure through
                stderr or non-zero return code, or if expected_fname check fails.
            WalltimeError: if the calculation has not completed before the end
                of the allotted walltime.

        Returns:
            (str, str): any output to stdout/stderr.

        """

        proc_clock = time.time()

        def _check_file_has_been_written(fname, proc_start):
            """ Check if the file at `fname` has been written to by the current process.
            Assumes a 1 second lee-way in the IO, in case the process wrote to the file
            just before starting the timer.

            Parameters:
                fname (str): the filename to check.
                proc_start (int): Unix time in seconds at which the process started.

            Raises:
                CalculationError: if the file does not exist, or if it is too old.

            """

            if not os.path.isfile(fname):
                msg = ('File {} was not created after {} s, please check your executable: {}.'
                       .format(fname, self.polltime, self.executable))
                LOG.critical(msg)
                raise CalculationError(msg)
            if os.path.getmtime(fname) - proc_clock < -1:
                msg = ('File {} present, but too old to be made by this process. '
                       'Please check your executable: {}.'
                       .format(fname, self.executable))
                LOG.critical(msg)
                raise CalculationError(msg)

        if expected_fname is not None:
            LOG.info("Watching for file write to {} after {} s".format(expected_fname, self.polltime))
            while process.poll() is None:
                proc_elapsed = time.time() - proc_clock
                if proc_elapsed > self.polltime:
                    _check_file_has_been_written(expected_fname, proc_clock)
                    break
                time.sleep(1)

        timeout = None
        if check_walltime and self.max_walltime is not None:
            LOG.info("Will not let process walltime exceed {} seconds".format(self.max_walltime))
            proc_elapsed = time.time() - proc_clock
            timeout = max(self.max_walltime - 2 * self.polltime - proc_elapsed, 1)

        try:
            out, errs = process.communicate(timeout=timeout)

            out = out.decode("utf-8")
            errs = errs.decode("utf-8")
            if process.returncode != 0 or errs:
                # as there are several reasons why the process can return != 0, handle
                # them in turn for each case, rather than raising an error here
                LOG.warning('Process returned error code {}'.format(process.returncode))
                LOG.warning('\nstdout: {}'.format(out))
                LOG.warning('\nstderr: {}'.format(errs))

        except sp.TimeoutExpired:
            LOG.error("Process reached maximum walltime, cleaning up...")
            self._times_up(process)
            raise WalltimeError("Cleaned up process after reaching maximum walltime")

        except Exception as err:
            LOG.error('Unexpected Exception {} caught: terminating job for {}.'.format(type(err).__name__, self.seed))
            raise err

        finally:
            LOG.error("Explicitly terminating process.")
            process.terminate()

        return out, errs

    def _catch_castep_errors(self, process):
        """ Look for CASTEP error files and fallover appropriately. If
        the magic string 'Work-around was succesful' is found in error,
        no errors will be reported and the file will be deleted,
        provided no other errors exist.
            results_dict (dict/str): dictionary containing data to u

        Returns:
            (bool, str, function): True if error files were found, otherwise False,
                followed by the error messages. If the error can be remedied,
                return the function to attempt to fix it.

        """
        err_file = '{}*err'.format(self.seed)
        error_str = ''
        errors_present = False
        remedy = None

        if process.returncode != 0:
            msg = 'CASTEP returned non-zero error code {}.'.format(process.returncode)
            error_str += msg + '\n'
            errors_present = True

        for globbed in glob.glob(err_file):
            if globbed.endswith('opt_err'):
                continue
            if os.path.isfile(globbed):
                with open(globbed, 'r') as f:
                    flines = f.readlines()

                for line in flines:
                    if 'Work-around was successful, continuing with calculation.' in line:
                        LOG.info('Found LAPACK issue that was circumvented, removing error file.')
                        os.remove(globbed)
                        break
                    elif 'ERROR in cell constraints: attempt to fix' in line:
                        remedy = self._remedy_castep_symmetry_error
                        break
                    elif 'Symmetry matrix not integer' in line:
                        remedy = self._remedy_symmetry_matrix_not_integer
                        break
                else:
                    error_str += ' '.join(flines)
                    error_str += '\n'
                    errors_present = True

        return errors_present, error_str, remedy

    @staticmethod
    def _remedy_castep_symmetry_error(opti_dict):
        """ Remedy the common cell constrains error by
        disabling symmetry.

        Parameters:
            opti_dict (dict): the dictionary of parameters to change.

        """
        LOG.info('Trying to remedy CASTEP symmetry error...')
        if 'symmetry_generate' in opti_dict:
            del opti_dict['symmetry_generate']
        if 'symmetry_tol' in opti_dict:
            del opti_dict['symmetry_tol']
        if 'snap_to_symmetry' in opti_dict:
            del opti_dict['snap_to_symmetry']

    def _remedy_symmetry_matrix_not_integer(self, opti_dict):
        LOG.info('Trying to remedy symmetry matrix not integer bug: symmetry tol now = {}'
                 .format(opti_dict['symmetry_tol']))
        if 'symmetry_tol' in opti_dict:
            opti_dict['symmetry_tol'] /= 10

    def mv_to_bad(self, seed):
        """ Move all files associated with "seed" to bad_castep, from both
        the compute directory (if it exists) and the root dir..

        Parameters:
            seed (str): filename of structure.

        """
        try:
            bad_dir = self.root_folder + '/' + self.paths["failed_dir"]
            if not os.path.exists(bad_dir):
                os.makedirs(bad_dir, exist_ok=True)
            seed_files = glob.glob(seed + '.*') + glob.glob(seed + '-out.cell*')
            if seed_files:
                LOG.info('Moving files to bad_castep: {bad}.'.format(bad=bad_dir))
                LOG.debug('Files to move: {seed}'.format(seed=seed_files))
                for _file in seed_files:
                    try:
                        shutil.copy2(_file, bad_dir)
                        os.remove(_file)
                    except Exception as exc:
                        LOG.warning('Error moving files to bad: {error}'.format(error=exc))
            # check root folder for any matching files and remove them
            fname = '{}/{}'.format(self.root_folder, seed)
            for ext in ['.res', '.res.lock', '.castep', '-out.cell']:
                if os.path.isfile('{}{}'.format(fname, ext)):
                    os.remove('{}{}'.format(fname, ext))
        except Exception as exc:
            LOG.warning('Error moving files to bad: {error}'.format(error=exc))

    def mv_to_completed(self, seed, completed_dir='completed', keep=False, skip_existing=False):
        """ Move all associated files to completed, removing any
        remaining files in the root_folder and compute_dir.

        Parameters:
            seed (str): filename for structure.

        Keyword arguments:
            completed_dir (str): folder for completed jobs.
            keep (bool): whether to also move intermediate files.
            skip_existing (bool): if True, skip files that already exist,
                otherwise throw an error.

        """
        completed_dir = self.root_folder + '/' + completed_dir
        LOG.info('Moving files to completed: {completed}'.format(completed=completed_dir))

        if seed.endswith('.res'):
            seed = str(seed.replace('.res', ''))

        if not os.path.exists(completed_dir):
            os.makedirs(completed_dir, exist_ok=True)

        for _file in glob.glob(seed + '*_bak') + glob.glob(seed + '*.lock'):
            os.remove(_file)
        if keep:
            seed_files = glob.glob(seed + '.*') + glob.glob(seed + '-out.cell*')
            if seed_files:
                LOG.debug('Files to move: {files}.'.format(files=seed_files))
                for _file in seed_files:
                    if skip_existing and os.path.isfile(completed_dir + '/' + _file):
                        LOG.warning('File already found {}...'.format(_file))
                    else:
                        shutil.move(_file, completed_dir)
        else:
            # move castep/param/res/out_cell files to completed
            file_exts = ['.castep']
            if self.kpts_1D:
                file_exts.append('.param')
            if not self.conv_kpt_bool and not self.conv_cutoff_bool:
                file_exts.append('.res')
            if os.path.isfile(seed + '-out.cell'):
                file_exts.append('-out.cell')
            if self.calc_doc.get('write_formatted_density'):
                file_exts.append('.den_fmt')
            for ext in file_exts:
                try:
                    shutil.move('{}{}'.format(seed, ext), completed_dir)
                except Exception as exc:
                    LOG.warning('Error moving files to completed: {error}'.format(error=exc))

        # delete whatever is left
        wildcard_fnames = glob.glob(seed + '.*')
        wildcard_fnames += glob.glob(seed + '-out.*')
        for fname in wildcard_fnames:
            os.remove(fname)

    def cp_to_input(self, seed, ext='res', glob_files=False):
        """ Copy initial cell and res to input folder.

        Parameters:
            seed (str): filename of structure.

        Keyword arguments:
            ext (str): file extension for structure.
            glob_files (bool): whether to glob all related seed files.

        """
        if not self._first_run:
            return
        input_dir = self.root_folder + '/input'
        LOG.debug('Copying file to input_dir: {input}'.format(input=input_dir))
        if not os.path.exists(input_dir):
            os.makedirs(input_dir, exist_ok=True)
        if glob_files:
            files = glob.glob('{}.*'.format(seed))
            LOG.debug('Files to copy: {files}'.format(files=files))
            for f in files:
                if f.endswith('.lock'):
                    continue
                if not os.path.isfile(input_dir + '/' + f):
                    shutil.copy2('{}'.format(f), input_dir)
        else:
            LOG.debug('File to copy: {file}'.format(file='{}.{}'.format(seed, ext)))
            if os.path.isfile('{}.{}'.format(seed, ext)):
                if not os.path.isfile('{}/{}.{}'.format(input_dir, seed, ext)):
                    shutil.copy2('{}.{}'.format(seed, ext), input_dir)

    def _setup_relaxation(self):
        """ Set up directories and files for relaxation.

        Raises:
            CalculationError: if structure has already exceeded
                geom_max_iter.
            CriticalError: if unable to split up calculation, normally
                indicating geom_max_iter is too small.

        """

        LOG.info('Preparing to relax {seed}'.format(seed=self.seed))
        self._setup_compute_dir(self.seed, self.compute_dir, custom_params=self.custom_params)

        # update res file with intermediate calculation if castep file is newer than res
        if os.path.isfile(self.seed + '.castep') and os.path.isfile(self.seed + '.res'):
            LOG.info('Trying to update res file with result from intermediate CASTEP file found in root_dir')
            if self.compute_dir is not None:
                shutil.copy2(self.seed + '.castep', self.compute_dir)
            castep_dict, success = castep2dict(self.seed + '.castep', db=False, verbosity=self.verbosity)
            if success:
                self.res_dict['geom_iter'] = castep_dict.get('geom_iter', 0)

            if os.path.getmtime(self.seed + '.res') < os.path.getmtime(self.seed + '.castep'):
                LOG.info('CASTEP file was updated more recently than res file, using intermediate structure...')
                self.res_dict.update(castep_dict)

        if self.compute_dir is not None:
            os.chdir(self.compute_dir)

        # copy initial res file to seed
        LOG.info('Writing fresh res file to start calculation from.')
        doc2res(self.res_dict, self.seed, info=False, hash_dupe=False, overwrite=True)
        self.cp_to_input(self.seed)

        # set up geom opt iteration options based on input/scraped parameters
        self._max_iter = self.calc_doc['geom_max_iter']
        if self.res_dict.get('geom_iter'):
            self._max_iter -= self.res_dict['geom_iter']
        else:
            self.res_dict['geom_iter'] = 0

        if self.res_dict['geom_iter'] > self.calc_doc['geom_max_iter']:
            msg = '{} iterations already performed on structure, exiting...'.format(self.res_dict['geom_iter'])
            LOG.critical(msg)
            raise CalculationError(msg)

        # number of steps in fine and rough calcs respectively
        fine_iter = self.fine_iter
        rough_iter = self.rough_iter
        # number of calculations to run with rough_iter steps
        num_rough_iter = self.rough
        if 'geom_method' in self.calc_doc:
            if self.calc_doc['geom_method'].lower() == 'tpsd' and rough_iter < 3:
                rough_iter = 3
        self._geom_max_iter_list = (num_rough_iter * [rough_iter])
        self._max_iter -= num_rough_iter * rough_iter
        if self.squeeze:
            self._squeeze_list = [True for val in self._geom_max_iter_list]
        else:
            self._squeeze_list = [False for val in self._geom_max_iter_list]

        num_fine_iter = ceil(int(self._max_iter) / fine_iter)
        if self._max_iter > 0:
            if self._max_iter < fine_iter:
                fine_iter = self._max_iter
                num_fine_iter = 1
            self._geom_max_iter_list.extend(num_fine_iter * [fine_iter])
            self._squeeze_list.extend(num_fine_iter * [False])

        LOG.info('Geometry optimisation iteration scheme set to {}'.format(self._geom_max_iter_list))
        if self.squeeze:
            LOG.info('Squeeze scheme set to {}'.format(self._geom_max_iter_list))

        if not self._geom_max_iter_list:
            msg = 'Could not divide up relaxation; consider increasing geom_max_iter'
            LOG.critical(msg)
            raise CriticalError(msg)

    def _update_input_files(self, seed, calc_doc, squeeze=None):
        """ Update the cell and param files for the next relaxation.

        Parameters:
            seed (str): the seedname to update.
            calc_doc (dict): the calculation dictionary to write to file.

        Keyword arguments:
            squeeze (float): external pressure to add this step

        """
        if seed is None:
            seed = self.seed

        this_calc_doc = deepcopy(calc_doc)

        # update cell
        if os.path.isfile(seed + '.cell'):
            os.remove(seed + '.cell')
        if squeeze is not None:
            if squeeze:
                LOG.info('Applying pressure of {} GPa to this calculation'.format(squeeze))
                this_calc_doc['external_pressure'] = [[squeeze, 0, 0], [0, squeeze, 0], [0, 0, squeeze]]
            else:
                LOG.info('Pressure reset to {}'.format(self._target_pressure))
                this_calc_doc['external_pressure'] = self._target_pressure

        if self.kpts_1D:
            n_kz = ceil(1 / (this_calc_doc['lattice_abc'][0][2] * self._target_spacing))
            if n_kz % 2 == 1:
                n_kz += 1
            this_calc_doc['kpoints_mp_grid'] = [1, 1, n_kz]
            if 'kpoints_mp_spacing' in calc_doc:
                del calc_doc['kpoints_mp_spacing']
        doc2cell(this_calc_doc, seed, hash_dupe=False, spin=self.spin)

        # update param
        if not self.custom_params:
            if os.path.isfile(seed + '.param'):
                os.remove(seed + '.param')
            doc2param(this_calc_doc, seed, hash_dupe=False, spin=self.spin)

        LOG.debug('Calculation dictionary: {}'
                  .format(matador.utils.print_utils.dumps(this_calc_doc,
                                                          indent=None)))

    def tidy_up(self, seed):
        """ Delete all created files before quitting.

        Parameters:
            seed (str): filename for structure.

        """
        files = glob.glob(seed + '.*')
        if self.compute_dir is not None:
            files += glob.glob(self.root_folder + '/' + seed + '.*')
        if files:
            LOG.info('Tidying up remaining files: {files}'.format(files=files))
            for f in files:
                # if we're working in a compute dir, then delete any remaining files in base dir
                # otherwise, only delete things that we dont care about, i.e. non-res/castep in
                # case they were not correctly moved to completed/bad_castep by the other routines
                if self.compute_dir is not None or (not (f.endswith('.res') or f.endswith('.castep'))):
                    os.remove(f)

    def _update_castep_output_files(self, seed, opti_dict=None):
        """ Copy new data to root CASTEP output files and update
         the results dict.

        Keyword arguments:
            opti_dict (dict): intermediate calculation results.

       """
        LOG.info('Updating .res and .castep files in root_dir with new results')
        if opti_dict is not None:
            if os.path.isfile(seed + '.res'):
                os.rename('{}.res'.format(seed), '{}.res_bak'.format(seed))
            try:
                doc2res(opti_dict, seed, hash_dupe=False)
            except CalculationError:
                doc2res(opti_dict, seed, hash_dupe=False, info=False)
            if os.path.isfile(seed + '.res_bak'):
                os.remove(seed + '.res_bak')
            self.res_dict.update(opti_dict)

        if self.compute_dir is not None:
            if os.path.isfile(seed + '.res'):
                shutil.copy2(seed + '.res', self.root_folder)
            if os.path.isfile(seed + '.castep'):
                shutil.copy2(seed + '.castep', self.root_folder)

    def _finalise_result(self, intermediate=False):
        """ Push to queue if necessary and return status.

        Keyword arguments:
            intermediate (bool): whether we want to run more calculations
                on the output of this, i.e. whether to move to completed
                or not.

        Returns:
            bool: True is relaxation was successful, False otherwise.

        """
        LOG.info('Finalising calculation...')
        try:
            success = self.res_dict.get('optimised', False)
        except AttributeError:
            success = False

        LOG.info('Was calculation successful? {success}'.format(success=success))
        if self.output_queue is not None:
            LOG.info('Pushing results to output queue')
            self.output_queue.put(self.res_dict)
        if success:
            if not intermediate:
                self.mv_to_completed(self.seed, completed_dir=self.paths['completed_dir'])
        else:
            self.mv_to_bad(self.seed)

        if success:
            self.final_result = self.res_dict

        if not intermediate:
            # clean up rest of files
            self.tidy_up(self.seed)

        return success

    def _times_up(self, process):
        """ If walltime has nearly expired, run this function
        to kill the process and unlock it for restarted calculations.

        Parameters:
            subprocess.Popen: running process to be killed.

        """
        if self.compute_dir is not None:
            LOG.info('Cleaning up compute_dir: {dir}'.format(dir=self.compute_dir))
            for f in glob.glob('{}.*'.format(self.seed)):
                shutil.copy2(f, self.root_folder)
                os.remove(f)
        LOG.info('Removing lock file so calculation can be continued.')
        if os.path.isfile('{}/{}{}'.format(self.root_folder, self.seed, '.res.lock')):
            os.remove('{}/{}{}'.format(self.root_folder, self.seed, '.res.lock'))

    @staticmethod
    def remove_compute_dir_if_finished(compute_dir):
        """ Delete the compute directory, provided it contains no
        calculation data.

        Parameters:
            compute_dir (str): path to compute directory.

        Returns:
            bool: True if folder was deleted as no res/castep files
                were found, otherwise False.

        """
        LOG.info('Checking if compute_dir still contains calculations...')
        if not os.path.isdir(compute_dir):
            return False

        files = glob.glob(compute_dir + '/*')
        LOG.debug('Found {files} in {dir}'.format(files=files, dir=compute_dir))

        for fname in files:
            if fname.endswith('.res') or fname.endswith('.castep'):
                LOG.debug('Not removing {dir} as it still contains calculation {fname}'.format(
                    dir=compute_dir, fname=fname))
                return False

        # remove files in directory, then delete directory
        LOG.debug('Deleting files {files} from {dir}'.format(files=files, dir=compute_dir))
        for fname in files:
            if os.path.isfile(fname):
                try:
                    os.remove(fname)
                except FileNotFoundError:
                    pass

        if os.path.isdir(compute_dir):
            LOG.debug('Deleting directory {dir}'.format(dir=compute_dir))
            try:
                os.rmdir(compute_dir)
            except OSError:
                LOG.debug('Unable to delete directory {} as it still contains files.'.format(compute_dir))

        if os.path.islink(compute_dir.split('/')[-1] + '_link'):
            os.remove(compute_dir.split('/')[-1] + '_link')

        return True

    def _setup_compute_dir(self, seed, compute_dir, custom_params=False, generic=False):
        """ Create the desired directory if it doens't exist,
        and try to link to it in the current folder.

        Parameters:
            seed (str): name of seed.
            compute_dir (str): name of directory to make.

        Keyword arguments:
            custom_params (bool): whether to try to copy custom
                param files into this directory.

        """
        if compute_dir is None:
            return

        LOG.info('Using compute_dir: {}'.format(compute_dir))
        if not os.path.isdir(compute_dir):
            try:
                os.makedirs(compute_dir)
            except PermissionError as exc:
                raise CriticalError('Invalid compute dir requested: {} {}'.format(compute_dir, exc))
        # if compute_dir isn't simply inside this folder, make a symlink that is
        if compute_dir.startswith('/'):
            link_name = compute_dir.split('/')[-1] + '_link'
            if not os.path.exists(link_name):
                os.symlink(compute_dir, link_name)

        if generic:
            # if generic, copy all seed files to compute dir
            for f in glob.glob(seed + '*'):
                shutil.copy2(f, compute_dir)

        # copy pspots and any intermediate calcs to compute_dir
        LOG.info('Copying pspots into compute_dir')
        if self.cell_dict is not None:
            for pspot in self.cell_dict.get('species_pot', {}).values():
                if os.path.isfile(pspot):
                    try:
                        shutil.copy2(pspot, compute_dir)
                    except shutil.SameFileError:
                        pass

        if custom_params:
            shutil.copy2(seed + '.param', compute_dir)

        # if a checkfile exists, copy it to the new dir
        # so that it can be restarted from
        if os.path.isfile(seed + '.check'):
            shutil.copy2(seed + '.check', compute_dir)

    def scf(self, *args, **kwargs):
        """ Alias for backwards-compatibility. """
        import warnings
        warnings.warn(
            '`scf` method name is deprecated, please use run_castep_singleshot to avoid future issues.',
            DeprecationWarning
        )
        return self.run_castep_singleshot(*args, **kwargs)

    def relax(self, *args, **kwargs):
        """ Alias for backwards-compatibility. """
        import warnings
        warnings.warn(
            '`relax` method name is deprecated, please use run_castep_relaxation to avoid future issues.',
            DeprecationWarning
        )
        return self.run_castep_relaxation(*args, **kwargs)



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements the scraper functions for the CIF
(Crystallographic Information File) format.

"""

import functools
from pathlib import Path
import numpy as np
from matador.scrapers.utils import scraper_function
from matador.utils.cell_utils import get_spacegroup_spg
from matador.utils.cell_utils import abc2cart, cart2volume, frac2cart

EPS = 1e-13


@scraper_function
def cif2dict(seed, **kwargs):
    """ Extract available information from  .cif file and store as a
    dictionary. Raw cif data is stored under the `'_cif'` key. Symmetric
    sites are expanded by the symmetry operations and their occupancies
    are tracked.

    Parameters:
        seed (str/list): filename or list of filenames of .cif file(s)
            (with or without extension).

    Returns:
        (dict/str, bool): if successful, a dictionary containing scraped
            data and True, if not, then an error string and False.


    """

    with open(seed, 'r') as f:
        flines = f.readlines()

    doc = dict()
    cif_dict = _cif_parse_raw(flines)

    doc['_cif'] = cif_dict
    doc['source'] = [str(Path(seed).resolve())]

    doc['atom_types'] = []
    atom_labels = cif_dict.get("_atom_site_type_symbol", False)
    if not atom_labels:
        atom_labels = cif_dict.get("_atom_site_label", False)
    if not atom_labels:
        raise RuntimeError(f"Unable to find atom types in cif file {seed}.")

    for atom in atom_labels:
        symbol = ''
        for character in atom:
            if not character.isalpha():
                break
            else:
                symbol += character
        doc['atom_types'].append(symbol)

    doc['positions_frac'] = [list(map(lambda x: float(x.split('(')[0]), vector)) for vector in
                             zip(cif_dict['_atom_site_fract_x'],
                                 cif_dict['_atom_site_fract_y'],
                                 cif_dict['_atom_site_fract_z'])]

    if '_atom_site_occupancy' in cif_dict:
        doc['site_occupancy'] = [float(x.split('(')[0]) for x in cif_dict['_atom_site_occupancy']]
    else:
        doc['site_occupancy'] = [1.0 for _ in doc['positions_frac']]

    if '_atom_site_symmetry_multiplicity' in cif_dict:
        doc['site_multiplicity'] = [float(x.split('(')[0]) for x in cif_dict['_atom_site_symmetry_multiplicity']]
    else:
        doc['site_multiplicity'] = [1.0 for _ in doc['positions_frac']]

    doc['lattice_abc'] = [list(map(_cif_parse_float_with_errors,
                                   [cif_dict['_cell_length_a'],
                                    cif_dict['_cell_length_b'],
                                    cif_dict['_cell_length_c']])),
                          list(map(_cif_parse_float_with_errors,
                                   [cif_dict['_cell_angle_alpha'],
                                    cif_dict['_cell_angle_beta'],
                                    cif_dict['_cell_angle_gamma']]))]

    doc['lattice_cart'] = abc2cart(doc['lattice_abc'])
    doc['cell_volume'] = cart2volume(doc['lattice_cart'])
    doc['stoichiometry'] = _cif_disordered_stoichiometry(doc)
    doc['num_atoms'] = len(doc['positions_frac'])

    if '_space_group_symop_operation_xyz' in doc['_cif'] and '_symmetry_equiv_pos_as_xyz' not in doc['_cif']:
        doc["_cif"]["_symmetry_equiv_pos_as_xyz"] = doc["_cif"]["_space_group_symop_operation_xyz"]
    if '_symmetry_equiv_pos_as_xyz' in doc['_cif']:
        _cif_set_unreduced_sites(doc)

    try:
        doc['space_group'] = get_spacegroup_spg(doc, check_occ=False)
    except RuntimeError:
        pass

    return doc, True


def _cif_parse_float_with_errors(x):
    """ Strip bracketed errors from end of float. """
    return float(x.split('(')[0])


def _cif_disordered_stoichiometry(doc):
    """ Create a matador stoichiometry normalised to the smallest integer
    number of atoms, unless all occupancies are 1/0.

    Parameters:
        doc: dictionary containing `atom_types`, `site_occupancy` and
            `site_multiplicity` keys.

    Returns:
        list of tuples: a standard matador stoichiometry.

    """
    from collections import defaultdict
    stoich = defaultdict(float)
    eps = 1e-8
    disordered = False
    for ind, site in enumerate(doc['atom_types']):
        stoich[site] += doc['site_occupancy'][ind] * doc['site_multiplicity'][ind]
        if doc['site_multiplicity'][ind] % 1 > 1e-5:
            disordered = True

    if disordered:
        min_int = 1e10
        for atom in stoich:
            if abs(int(stoich[atom]) - stoich[atom]) < eps:
                if int(stoich[atom]) < min_int:
                    min_int = int(stoich[atom])

        if min_int == 1e10:
            min_int = 1
        for atom in stoich:
            stoich[atom] /= min_int

    return sorted([[atom, stoich[atom]] for atom in stoich])


def _cif_parse_raw(flines):
    """ Parse raw CIF file data into a dictionary.

    Parameters:
        flines (:obj:`list` of :obj:`str`): contents of .cif file.

    Returns:
        dict: dictionary containing cif data with native fields/ordering.

    """
    ind = 0
    cif_dict = dict()
    cif_dict['loops'] = list()
    while ind < len(flines):
        jnd = 1
        line = flines[ind].strip()
        # parse single (multi-line) tag
        if line.startswith('_'):
            line = line.split()
            key = line[0]
            data = ''
            if len(line) > 1:
                data += ' '.join(line[1:])
            while ind + jnd < len(flines) and _cif_line_contains_data(flines[ind+jnd].strip()):
                data += flines[ind+jnd].strip().replace(';', '')
                jnd += 1
            cif_dict[key] = data.strip()
        # parse loop block
        elif line.startswith('loop_'):
            # get loop keys
            keys = []
            while flines[ind+jnd].strip().startswith('_'):
                keys.append(flines[ind+jnd].strip())
                jnd += 1
            for key in keys:
                cif_dict[key] = []
            cif_dict['loops'].append(keys)
            while ind + jnd < len(flines) and _cif_line_contains_data(flines[ind+jnd].strip()):
                data = ''
                # loop over line and next lines
                while ind + jnd < len(flines) and _cif_line_contains_data(flines[ind+jnd]):
                    data += flines[ind+jnd]
                    jnd += 1

            loop_dict = _cif_parse_loop(keys, data)
            cif_dict.update(loop_dict)

        ind += jnd

    return cif_dict


def _cif_parse_loop(keys, data_block):
    """ A hacky way to parse CIF data loops that can be split by quotes
    or spaces. There must be a better way...

    Parameters:
        keys (list of str): list of keys for the loop.
        data_block (str): raw string of the entire data block.

    Returns:
        Dict[str, str]: a dictionary with keys from ``keys``, containing the
            data split by quotes and spaces. All data is left as
            strings for further processing.

    """

    from collections import deque, defaultdict

    dq = deque(data_block)
    data_list = []
    entry = None
    in_quotes = False
    while dq:
        char = dq.popleft()
        if not char.strip() and entry is None:
            continue
        elif (not char.strip() or char in [" ", ";"]) and entry is not None and not in_quotes:
            data_list.append(entry.strip())
            entry = None
        elif not char.strip() and entry is not None and in_quotes:
            entry += " "
        elif char == "'" and entry and entry is not None:
            in_quotes = False
            data_list.append(entry.strip())
            entry = None
        elif char == "'" and entry is None:
            entry = ''
            in_quotes = True
        else:
            if entry is None:
                entry = char
            else:
                entry += char

    loop_dict = defaultdict(list)
    for ind, entry in enumerate(data_list):
        ind = ind % len(keys)
        loop_dict[keys[ind]].append(entry)

    return loop_dict


def _cif_set_unreduced_sites(doc):
    """ Expands sites by symmetry operations found under the key
    `symemtry_equiv_pos_as_xyz` in the cif_dict.

    Parameters:
        doc (dict): matador document to modify. Must contain symops
            under doc['_cif']['_symmetry_equiv_pos_as_xyz']. This doc
            is updated with new `positions_frac`, `num_atoms`, `atom_types`
            and `site_occupancy`.

    """
    from matador.utils.cell_utils import wrap_frac_coords
    from matador.utils.cell_utils import calc_pairwise_distances_pbc
    from matador.fingerprints.pdf import PDF

    species_sites = dict()
    species_occ = dict()

    symmetry_ops = []
    symmetry_functions = []

    def _apply_sym_op(x=None, y=None, z=None, symmetry=None):
        """ Returns the site after the applied symmetry operation, in string representation. """
        # cannot use a listcomp here due to interplay with functools
        return [eval(symmetry[0]), eval(symmetry[1]), eval(symmetry[2])]

    for symmetry in doc['_cif']['_symmetry_equiv_pos_as_xyz']:
        symmetry = tuple(elem.strip() for elem in symmetry.strip('\'').split(','))
        # check the element before doing an eval, as it is so unsafe
        allowed_chars = ['x', 'y', 'z', '.', '/', '+', '-',
                         '0', '1', '2', '3', '4', '5', '6', '7', '8', '9']
        for element in symmetry:
            for character in element:
                if character not in allowed_chars:
                    raise RuntimeError('You are trying to do something naughty with the symmetry element {}'
                                       .format(element))
        symmetry_ops.append(symmetry)
        symmetry_functions.append(functools.partial(_apply_sym_op, symmetry=symmetry))

    for ind, site in enumerate(doc['positions_frac']):
        species = doc['atom_types'][ind]
        occupancy = doc['site_occupancy'][ind]
        if doc['atom_types'][ind] not in species_sites:
            species_sites[species] = []
            species_occ[species] = []
        for symmetry in symmetry_functions:
            x, y, z = site
            new_site = symmetry(x=x, y=y, z=z)
            new_site = wrap_frac_coords([new_site])[0]
            species_sites[species].append(new_site)
            species_occ[species].append(occupancy)

    unreduced_sites = []
    unreduced_occupancies = []
    unreduced_species = []

    # this loop assumes that no symmetry operation can map 2 unlike sites upon one another
    for species in species_sites:
        unreduced_sites.extend(species_sites[species])
        unreduced_occupancies.extend(species_occ[species])
        unreduced_species.extend(len(species_sites[species]) * [species])

    # check that the symmetry procedure has not generated overlapping atoms
    # this can happen for certain symmetries/cells if positions are not
    # reported to sufficient precision
    images = PDF._get_image_trans_vectors_auto(
        doc['lattice_cart'],
        0.1, 0.01, max_num_images=1,
    )

    poscarts = frac2cart(doc['lattice_cart'], unreduced_sites)
    distances = calc_pairwise_distances_pbc(
        poscarts,
        images,
        doc['lattice_cart'],
        0.01,
        compress=False,
        filter_zero=False,
        per_image=True
    )

    dupe_set = set()
    for img in distances:
        i_s, j_s = np.where(~img.mask)
        for i, j in zip(i_s, j_s):
            if i == j:
                continue
            else:
                # sites can overlap if they have partial occupancy
                if i not in dupe_set and unreduced_species[i] == unreduced_species[j]:
                    dupe_set.add(j)

    doc['positions_frac'] = unreduced_sites
    doc['site_occupancy'] = unreduced_occupancies
    doc['atom_types'] = unreduced_species

    doc['site_occupancy'] = [
        atom for ind, atom in enumerate(unreduced_occupancies) if ind not in dupe_set
    ]
    doc['atom_types'] = [
        atom for ind, atom in enumerate(unreduced_species) if ind not in dupe_set
    ]
    doc['positions_frac'] = [
        atom for ind, atom in enumerate(unreduced_sites) if ind not in dupe_set
    ]

    _num_atoms = np.sum(doc['site_occupancy'])
    if abs(_num_atoms - round(_num_atoms, 0)) < EPS:
        _num_atoms = int(round(_num_atoms, 0))
    doc['num_atoms'] = _num_atoms

    if len(doc['site_occupancy']) != len(doc['positions_frac']):
        raise RuntimeError('Size mismatch between positions and occs, {} vs {}'
                           .format(len(doc['site_occupancy']), len(doc['positions_frac'])))
    if len(doc['positions_frac']) != len(doc['atom_types']):
        raise RuntimeError('Size mismatch between positions and types')


def _cif_line_contains_data(line):
    """ Check if string contains cif-style data. """
    return not any([line.startswith('_'), line.startswith('#'), line.startswith('loop_')])


@scraper_function
def _ase_cif2dict(fname):
    """ Read cif file into ASE object,
    then convert ASE Atoms into matador document.

    Parameters:
        fname (str): cif filename

    Returns:
        (dict, bool): simple matador document with error status.

    """
    import ase.io
    from matador.utils.ase_utils import ase2dict
    fname = fname.replace('.cif', '')
    atoms = ase.io.read(fname + '.cif')
    doc = ase2dict(atoms)

    return doc, True



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements some useful wrappers to the print function for
writing errors and warnings to stderr.

"""


import sys
import json
import numpy as np


def print_warning(string):
    """ Print but angry. """
    print('\033[93m', end='', file=sys.stderr)
    print(string, end='', file=sys.stderr)
    print('\033[0m', file=sys.stderr)


def print_failure(string):
    """ Print but sad. """
    print('\033[91m\033[4m', end='', file=sys.stderr)
    print(string, end='', file=sys.stderr)
    print('\033[0m', file=sys.stderr)


def print_success(string):
    """ Print but happy. """
    print('\033[92m\033[1m', end='')
    print(string, end='')
    print('\033[0m')


def print_notify(string):
    """ Print but aloof. """
    print('\033[94m', end='')
    print(string, end='')
    print('\033[0m')


def dumps(obj, **kwargs):
    """ Mirrors `json.dumps` whilst handling numpy arrays. """
    return json.dumps(obj, cls=NumpyEncoder, **kwargs)


class NumpyEncoder(json.JSONEncoder):
    """ This encoder handles NumPy arrays in JSON, and was
    taken from StackOverflow (where else)
    (https://stackoverflow.com/a/47626762).

    """
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)



# coding: utf-8
# Distributed under the terms of the MIT License.

""" The fingerprints module provides functionality for calculating, in
parallel, structural fingerprints like PDF and PXRD.

"""


__all__ = ['get_uniq_cursor',
           'Fingerprint', 'FingerprintFactory',
           'PDF', 'PDFOverlap', 'CombinedProjectedPDF',
           'PXRD']

__author__ = 'Matthew Evans'
__maintainer__ = 'Matthew Evans'


from matador.fingerprints.similarity import get_uniq_cursor
from matador.fingerprints.fingerprint import Fingerprint, FingerprintFactory
from matador.fingerprints.pdf import PDF, PDFOverlap, CombinedProjectedPDF
from matador.fingerprints.pxrd import PXRD



import unittest


class ScatteringDataLoaderTests(unittest.TestCase):
    """ Test whether the atomic scattering factors can be loaded. """

    def test_load_gsas(self):
        from matador.data import GSAS_ATOMIC_SCATTERING_COEFFS

        self.assertEqual(len(GSAS_ATOMIC_SCATTERING_COEFFS["Cu"]), 3)
        for key in GSAS_ATOMIC_SCATTERING_COEFFS:
            self.assertEqual(len(GSAS_ATOMIC_SCATTERING_COEFFS[key]), 3)

    def test_load_raspa(self):
        from matador.data import RASPA_ATOMIC_SCATTERING_COEFFS

        self.assertEqual(len(RASPA_ATOMIC_SCATTERING_COEFFS["Cu"]), 3)
        for key in RASPA_ATOMIC_SCATTERING_COEFFS:
            self.assertEqual(len(RASPA_ATOMIC_SCATTERING_COEFFS[key]), 3)



""" This submodule implements functions useful for plotting
the results of infrared and Raman spectroscopy calculations.

"""

import numpy as np
from matador.scrapers import phonon2dict
from matador.plotting.plotting import plotting_function
from matador.utils.chem_utils import INVERSE_CM_TO_EV


@plotting_function
def plot_ir_spectrum(seed, bin_width=1.0, ax=None, show=True, **kwargs):
    """ This function plots the IR/Raman spectrum found in the given .phonon file.

    Parameters:
        seed (str): the filename to scrape and plot.

    Keyword arguments:
        bin_width (float): the bin width for the IR plot.
        ax (matplotlib.axes.Axes): an existing axis on which to plot.
        show (bool): whether or not to display the plot in an X window

    Raises:
        RuntimeError: if unable to scrape IR data.

    Returns:
        matplotlib.axes.Axes: matplotlib axis with plotted data.

    """

    import matplotlib.pyplot as plt

    if ax is None:
        fig, ax_ir = plt.subplots(figsize=kwargs.get('figsize'))
    else:
        ax_ir = ax

    ir_data, s = phonon2dict(seed)
    if not s or 'infrared_intensity' not in ir_data:
        raise RuntimeError('Error scraping file: no IR intensities. {}'.format(ir_data))

    wavenumbers = ir_data['eigenvalues_q'] / INVERSE_CM_TO_EV

    plotting_raman = 'raman_intensity' in ir_data

    max_wavenumber = np.max(wavenumbers)
    bins = np.linspace(0, 1.2 * max_wavenumber, num=int((1 / bin_width) * 1.2 * max_wavenumber))

    ir_spectrum, bin_edges = np.histogram(
        wavenumbers,
        bins=bins,
        weights=ir_data['infrared_intensity'],
        density=True
    )
    # normalize so the highest peak is at 1
    ir_spectrum /= np.max(ir_spectrum)
    if plotting_raman:
        raman_spectrum, _ = np.histogram(
            wavenumbers,
            bins=bins,
            weights=ir_data['raman_intensity'],
            density=True
        )
        raman_spectrum /= np.max(raman_spectrum)

    # get bin centres
    bins = 0.5 * (bin_edges[1:] + bin_edges[:-1])

    ax_ir.plot(bins, ir_spectrum, color='#EE3425')
    ax_ir.set_xlabel('Wavenumber (cm$^{-1}$)')
    ax_ir.set_ylabel('Relative IR intensity', color='#EE3425')
    plt.gca().invert_yaxis()
    plt.gca().invert_xaxis()
    ir_max = np.max(ir_spectrum)
    ax_ir.set_ylim(2.1 * ir_max, -ir_max * 0.05)
    ax_ir.set_yticks(np.linspace(0, 1, 6))

    if plotting_raman:
        ax_raman = ax_ir.twinx()
        raman_max = np.max(raman_spectrum)
        ax_raman.set_yticks(np.linspace(0, 1, 6))
        ax_raman.set_ylim(-raman_max * 0.05, 2.1 * raman_max)
        ax_raman.plot(bins, raman_spectrum, color='#236DE8')
        ax_raman.set_ylabel('Relative Raman activity', color='#236DE8')

    plt.title(seed)

    if any([kwargs.get('pdf'), kwargs.get('svg'), kwargs.get('png')]):
        filename = seed.split('/')[-1].replace('.phonon', '') + '_ir'
        if kwargs.get('pdf'):
            plt.savefig('{}.pdf'.format(filename),
                        bbox_inches='tight', transparent=True)
        if kwargs.get('svg'):
            plt.savefig('{}.svg'.format(filename),
                        bbox_inches='tight', transparent=True)
        if kwargs.get('png'):
            plt.savefig('{}.png'.format(filename),
                        bbox_inches='tight', transparent=True)

    elif show:
        plt.show()

    return ax



#!/usr/bin/env python
# standard library
import unittest
from os.path import realpath

# matador modules
from matador.orm.orm import DataContainer

# grab abs path for accessing test data
REAL_PATH = "/".join(realpath(__file__).split("/")[:-1]) + "/"


class ORMTest(unittest.TestCase):
    def test_data_container(self):
        data_dict = {"a": "b", "lattice_abc": [[10, 10, 10], [90, 90, 90]]}
        container = DataContainer(data_dict)
        self.assertEqual(container["a"], "b")
        data_dict["a"] = 2
        self.assertEqual(container["a"], "b")

        self.assertEqual(container["lattice_abc"], [[10, 10, 10], [90, 90, 90]])
        data_dict["lattice_abc"][0][0] = 100
        self.assertEqual(container["lattice_abc"], [[10, 10, 10], [90, 90, 90]])

        with self.assertRaises(AttributeError):
            container["lattice_abc"] = [[1, 2, 3], [4, 5, 6]]

        container["extra_data"] = [1, 2, 3]



#!/usr/bin/env python
import unittest
from matador.fingerprints.pdf import PDF, PDFOverlap, PDFFactory
from matador.scrapers.castep_scrapers import res2dict
from matador.utils.cell_utils import abc2cart, cart2volume
import numpy as np
from os.path import realpath

REAL_PATH = "/".join(realpath(__file__).split("/")[:-1]) + "/"
DEBUG = True


class PDFCalculatorTest(unittest.TestCase):
    """ Test PDF calculator. """

    def test_auto_images_vs_large(self):
        doc, success = res2dict(REAL_PATH + "data/LiPZn-r57des.res")
        doc["lattice_cart"] = abc2cart(doc["lattice_abc"])
        doc["text_id"] = ["pdf", "test"]
        doc["pdf_num_images"] = PDF(
            doc,
            low_mem=True,
            num_images=5,
            rmax=15,
            dr=0.1,
            **{"debug": True, "projected": False}
        )
        doc["pdf_auto_images"] = PDF(
            doc,
            low_mem=True,
            num_images="auto",
            rmax=15,
            dr=0.1,
            **{"debug": True, "projected": False}
        )
        np.testing.assert_array_almost_equal(
            doc["pdf_num_images"].gr, doc["pdf_auto_images"].gr
        )

    def test_pdf_from_projected(self):
        doc, success = res2dict(REAL_PATH + "data/LiPZn-r57des.res")
        doc["lattice_cart"] = abc2cart(doc["lattice_abc"])
        doc["text_id"] = ["unprojected", "test"]
        doc["pdf_unprojected"] = PDF(doc, dr=0.01, **{"debug": False})
        doc["text_id"] = ["projected", "test"]
        doc["pdf_projected"] = PDF(doc, dr=0.01, **{"debug": False})
        np.testing.assert_array_almost_equal(
            doc["pdf_unprojected"].gr, doc["pdf_projected"].gr
        )

    def test_identity_overlap(self):
        doc, success = res2dict(REAL_PATH + "data/LiPZn-r57des.res")
        doc["lattice_cart"] = abc2cart(doc["lattice_abc"])
        doc["text_id"] = ["pdf", "test"]
        doc["pdf_smear"] = PDF(
            doc,
            num_images=3,
            dr=0.001,
            gaussian_width=0.1,
            style="smear",
            debug=False,
            low_mem=True,
        )
        overlap = PDFOverlap(doc["pdf_smear"], doc["pdf_smear"])
        self.assertEqual(overlap.similarity_distance, 0.0)

    @unittest.skip
    def test_overlap_smear_vs_hist(self):
        doc, success = res2dict(REAL_PATH + "data/LiPZn-r57des.res")
        doc["lattice_cart"] = abc2cart(doc["lattice_abc"])
        doc["text_id"] = ["smear", "test"]
        doc["pdf_smear"] = PDF(
            doc,
            num_images=3,
            dr=0.01,
            gaussian_width=0.1,
            projected=False,
            style="smear",
            low_mem=True,
        )
        doc["text_id"] = ["hist", "test"]
        doc["pdf_hist"] = PDF(
            doc, num_images=3, dr=0.1, projected=False, style="histogram"
        )
        overlap = PDFOverlap(doc["pdf_smear"], doc["pdf_hist"])
        self.assertLessEqual(overlap.similarity_distance, 0.02)
        self.assertGreater(overlap.similarity_distance, 0.0)

    def test_pdf_primitive_vs_supercell(self):
        test_doc, success = res2dict(REAL_PATH + "data/KP_primitive.res", db=False)
        test_doc["text_id"] = ["primitive", "cell"]
        test_doc["lattice_cart"] = abc2cart(test_doc["lattice_abc"])
        test_doc["cell_volume"] = cart2volume(test_doc["lattice_cart"])
        supercell_doc, success = res2dict(REAL_PATH + "data/KP_supercell.res", db=False)
        supercell_doc["text_id"] = ["supercell", "cell"]
        supercell_doc["lattice_cart"] = abc2cart(supercell_doc["lattice_abc"])
        supercell_doc["cell_volume"] = cart2volume(supercell_doc["lattice_cart"])
        test_doc["pdf"] = PDF(
            test_doc, dr=0.01, low_mem=True, rmax=10, num_images="auto", debug=DEBUG
        )
        supercell_doc["pdf"] = PDF(
            supercell_doc,
            dr=0.01,
            low_mem=True,
            rmax=10,
            num_images="auto",
            debug=DEBUG,
        )
        overlap = PDFOverlap(test_doc["pdf"], supercell_doc["pdf"])
        self.assertLessEqual(overlap.similarity_distance, 1e-3)
        self.assertGreaterEqual(overlap.similarity_distance, 0.0)

    @unittest.skip("Not sure this test is worthwhile anymore...")
    def test_ideal_gas_pdf(self, retry=0):
        """ DEPRECATED.

        Slow, and not very useful.

        """
        # create fake matador doc
        doc = dict()
        max_retries = 1
        self.assertLess(
            retry, max_retries, msg="After {} attempts, PDF still failed.".format(retry)
        )
        num_atoms = 100
        box_size = 10
        num_samples = 1
        rmax = 15
        dr = 0.01
        num_images = "auto"
        i = 0
        doc["atom_types"] = num_atoms * ["C"]
        doc["lattice_cart"] = np.asarray(
            [[box_size, 0, 0], [0, box_size, 0], [0, 0, box_size]]
        )
        doc["cell_volume"] = box_size ** 3
        doc["text_id"] = ["ideal", "gas"]
        while i < num_samples:
            doc["positions_frac"] = np.random.rand(num_atoms, 3)
            doc["text_id"] = "hist"
            doc["pdf"] = PDF(
                doc,
                num_images=num_images,
                dr=dr,
                rmax=rmax,
                lazy=True,
                style="histogram",
                debug=True,
            )
            doc["pdf"].calc_pdf()
            self.assertAlmostEqual(np.mean(doc["pdf"].gr[50:]), 1.0, places=1)

    def test_single_atom_pdf(self):
        from math import ceil

        doc = dict()
        box_size = 20
        rmax = 41
        dr = 0.1
        num_images = "auto"
        doc["positions_frac"] = [[0.5, 0.5, 0.5]]
        doc["atom_types"] = ["C"]
        doc["lattice_cart"] = np.asarray(
            [[box_size, 0, 0], [0, box_size, 0], [0, 0, box_size]]
        )
        doc["cell_volume"] = box_size ** 3
        doc["text_id"] = ["hist", "ogram"]
        doc["pdf"] = PDF(
            doc,
            num_images=num_images,
            dr=dr,
            rmax=rmax,
            lazy=True,
            style="histogram",
            debug=True,
        )
        doc["pdf"].calc_pdf()
        doc["text_id"] = ["smear"]
        doc["pdf_smear"] = PDF(
            doc,
            num_images=num_images,
            gaussian_width=0.01,
            dr=0.1,
            rmax=rmax,
            lazy=True,
            style="smear",
            debug=True,
        )
        doc["pdf_smear"].calc_pdf()
        doc["text_id"] = ["low"]
        doc["pdf_low"] = PDF(
            doc,
            low_mem=True,
            num_images=num_images,
            gaussian_width=0.01,
            dr=0.1,
            rmax=rmax,
            lazy=True,
            style="smear",
            debug=True,
        )
        doc["pdf_low"].calc_pdf()
        peaks = [20, np.sqrt(2) * 20, np.sqrt(3) * 20, 40]
        indices = [ceil(peak / dr) for peak in peaks]
        self.assertListEqual(np.where(doc["pdf_low"].gr > 1e-8)[0].tolist(), indices)
        self.assertListEqual(np.where(doc["pdf"].gr > 1e-8)[0].tolist(), indices)
        self.assertListEqual(np.where(doc["pdf_smear"].gr > 1e-8)[0].tolist(), indices)


class PDFFactoryCalculatorTest(unittest.TestCase):
    """ Test PDFFactory. """

    def test_concurrent_pdf(self):
        import glob
        import numpy as np
        import time
        from copy import deepcopy

        files = glob.glob(REAL_PATH + "data/hull-KPSn-KP/*.res")[0:24]
        cursor = [res2dict(file, db=False)[0] for file in files]
        serial_cursor = deepcopy(cursor)
        pdf_args = {
            "dr": 0.1,
            "num_images": "auto",
            "gaussian_width": 0.1,
            "lazy": False,
            "projected": False,
        }
        start = time.time()
        pdf_factory = PDFFactory(cursor, **pdf_args)
        factory_elapsed = time.time() - start
        start = time.time()
        for doc in serial_cursor:
            doc["pdf"] = PDF(doc, **pdf_args, timing=False)
        serial_elapsed = time.time() - start
        print(
            "{:.2f} s over {} processes vs {:.2f} s in serial".format(
                factory_elapsed, pdf_factory.nprocs, serial_elapsed
            )
        )
        print(
            "Corresponding to a speedup of {:.1f} vs ideal {:.1f}".format(
                serial_elapsed / factory_elapsed, pdf_factory.nprocs
            )
        )
        for ind, doc in enumerate(serial_cursor):
            np.testing.assert_array_almost_equal(
                doc["pdf"].gr, cursor[ind]["pdf"].gr, decimal=6
            )

    def test_concurrent_pdf_stoichs(self):
        import glob
        import numpy as np
        import time
        from copy import deepcopy
        from matador.hull import QueryConvexHull

        files = glob.glob(REAL_PATH + "data/hull-KPSn-KP/*.res")
        cursor = [res2dict(file, db=True)[0] for file in files]
        hull = QueryConvexHull(
            cursor=cursor,
            no_plot=True,
            hull_cutoff=0.5,
            summary=True,
            elements=["K", "Sn", "P"],
            quiet=True,
        )
        serial_cursor = deepcopy(hull.cursor)

        pdf_args = {
            "dr": 0.1,
            "num_images": "auto",
            "gaussian_width": 0.1,
            "lazy": False,
            "projected": False,
        }
        start = time.time()
        pdf_factory = PDFFactory(hull.cursor, **pdf_args)
        factory_elapsed = time.time() - start
        start = time.time()
        for doc in serial_cursor:
            doc["pdf"] = PDF(doc, **pdf_args, timing=False)
        serial_elapsed = time.time() - start
        print(
            "{:.2f} s over {} processes vs {:.2f} s in serial".format(
                factory_elapsed, pdf_factory.nprocs, serial_elapsed
            )
        )
        print(
            "Corresponding to a speedup of {:.1f} vs ideal {:.1f}".format(
                serial_elapsed / factory_elapsed, pdf_factory.nprocs
            )
        )
        for ind, doc in enumerate(serial_cursor):
            np.testing.assert_array_almost_equal(
                doc["pdf"].gr, hull.cursor[ind]["pdf"].gr, decimal=6
            )

    def test_concurrent_projected_pdf(self):
        import glob
        import numpy as np
        import time
        from copy import deepcopy

        files = glob.glob(REAL_PATH + "data/hull-KPSn-KP/*.res")[0:24]
        cursor = [res2dict(file, db=False)[0] for file in files]
        serial_cursor = deepcopy(cursor)
        pdf_args = {
            "dr": 0.1,
            "num_images": "auto",
            "gaussian_width": 0.1,
            "lazy": False,
            "projected": True,
        }
        start = time.time()
        pdf_factory = PDFFactory(cursor, **pdf_args)
        factory_elapsed = time.time() - start
        start = time.time()
        for doc in serial_cursor:
            doc["pdf"] = PDF(doc, **pdf_args, timing=False)
        serial_elapsed = time.time() - start
        print(
            "{:.2f} s over {} processes vs {:.2f} s in serial".format(
                factory_elapsed, pdf_factory.nprocs, serial_elapsed
            )
        )
        print(
            "Corresponding to a speedup of {:.1f} vs ideal {:.1f}".format(
                serial_elapsed / factory_elapsed, pdf_factory.nprocs
            )
        )
        for ind, doc in enumerate(serial_cursor):
            np.testing.assert_array_almost_equal(
                doc["pdf"].gr, cursor[ind]["pdf"].gr, decimal=6
            )
            for key in doc["pdf"].elem_gr:
                np.testing.assert_array_almost_equal(
                    doc["pdf"].elem_gr[key], cursor[ind]["pdf"].elem_gr[key], decimal=6
                )


if __name__ == "__main__":
    unittest.main()



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements the PXRD class for simulating powder XRD pattern
of a crystal.

"""

import itertools
import os
from typing import Tuple

import numpy as np

from matador.fingerprints.fingerprint import Fingerprint, FingerprintFactory
from matador.crystal import Crystal
from matador.utils.cell_utils import standardize_doc_cell, real2recip
from matador.utils.chem_utils import get_formula_from_stoich

THETA_TOL = 1e-5


class PXRD(Fingerprint):
    """ This class for computes powder X-ray diffraction patterns  of a
    given crystal for a certain incident wavelength. The cell is
    standardised with spglib before computing PXRD.

    This calculation takes into account atomic scattering factors, Lorentz
    polarisation and thermal broadening (with Debye-Waller factors set to
    1). Note: this class does not perform any q-dependent peak broadening,
    and instead uses a simple Lorentzian broadening. The default width
    of 0.03 provides good agreement with e.g. GSAS-II's default CuKa setup.
    Only one wavelength can be used at a time, but multiple patterns could be
    combined post hoc.

    Attributes:
        self.peak_positions (numpy.ndarray): sorted peak positions as
            values in 2θ
        self.hkls (numpy.ndarray): Miller indices correspnding to peaks,
            sorted by peak angle.
        self.peak_intensities (numpy.ndarray): intensity of each peak.
        self.pattern (numpy.ndarray): Lorentzian-broadened pattern at
            values of self.two_thetas.
        self.two_thetas (numpy.ndarray): continuous space of 2θ values
            corresponding to sample points of self.pattern.

    """
    def __init__(
        self,
        doc,
        wavelength: float = 1.5406,
        lorentzian_width: float = 0.03,
        two_theta_resolution: float = 0.01,
        two_theta_bounds: Tuple[float, float] = (0, 90),
        theta_m: float = 0.0,
        scattering_factors: str = "RASPA",
        lazy=False,
        plot=False,
        progress=False,
        *args,
        **kwargs
    ):
        """ Set up the PXRD, and compute it, if lazy is False.

        Parameters:
            doc (dict/Crystal): matador document to compute PXRD for.

        Keyword arguments:
            lorentzian_width (float): width of Lorentzians for broadening (DEFAULT: 0.03)
            wavelength (float): incident X-ray wavelength
                (DEFAULT: CuKa, 1.5406).
            theta_m (float): the monochromator angle in degrees (DEFAULT: 0)
            two_theta_resolution (float): resolution of grid 2θ
                used for plotting.
            two_theta_bounds (tuple of float): values between which
                to compute the PXRD pattern.
            scattering_factors (str): either "GSAS" or "RASPA" (default),
                which set of atomic scattering factors to use.
            lazy (bool): whether to compute PXRD or just set it up.
            plot (bool): whether to display PXRD as a plot.

        """
        self.wavelength = wavelength
        self.lorentzian_width = lorentzian_width
        self.two_theta_resolution = two_theta_resolution
        if two_theta_bounds is not None:
            self.two_theta_bounds = list(two_theta_bounds)
        else:
            self.two_theta_bounds = [0, 90]
        self.theta_m = theta_m
        self.scattering_factors = scattering_factors
        self.progress = progress

        if self.two_theta_bounds[0] < THETA_TOL:
            self.two_theta_bounds[0] = THETA_TOL

        if np.min(doc.get('site_occupancy', [1.0])) < 1.0:
            print("System has partial occupancy, not refining with spglib.")
            self.doc = Crystal(doc)
        else:
            self.doc = Crystal(standardize_doc_cell(doc, primitive=True))

        self.formula = get_formula_from_stoich(self.doc['stoichiometry'], tex=True)
        self.spg = self.doc['space_group']

        species = list(set(self.doc['atom_types']))

        # this could be cached across PXRD objects but is much faster than the XRD calculation itself
        if self.scattering_factors == "GSAS":
            from matador.data import GSAS_ATOMIC_SCATTERING_COEFFS
            self.atomic_scattering_coeffs = {spec: GSAS_ATOMIC_SCATTERING_COEFFS[spec] for spec in species}
        elif self.scattering_factors == "RASPA":
            from matador.data import RASPA_ATOMIC_SCATTERING_COEFFS
            self.atomic_scattering_coeffs = {spec: RASPA_ATOMIC_SCATTERING_COEFFS[spec] for spec in species}
        else:
            raise RuntimeError(
                "No set of scattering factors matched: {}. Please use 'GSAS' or 'RASPA'."
                .format(self.scattering_factors)
            )

        if not lazy:
            self.calculate()
            if plot:
                self.plot()

    def calc_pxrd(self):
        """ Calculate the PXRD pattern. """

        # set crystallographic data
        lattice_abc = np.asarray(self.doc.lattice_abc)
        lattice_cart = np.asarray(self.doc.lattice_cart)
        positions_abs = np.asarray(self.doc.positions_abs)
        site_occupancies = np.asarray(self.doc.site_occupancies)

        # find allowed reciprocal lattice points within limiting sphere
        min_r, max_r = [2 / self.wavelength * np.sin(np.pi / 180 * t / 2) for t in self.two_theta_bounds]
        Ns = np.floor(max_r * lattice_abc[0, :]).astype(int)

        recip = np.asarray(real2recip(lattice_cart)).T
        qs = np.zeros((2*sum(Ns), 3), dtype=np.float64)
        hkls = np.asarray(
            list(itertools.product(
                range(-Ns[0], Ns[0] + 1),
                range(-Ns[1], Ns[1] + 1),
                range(-Ns[2], Ns[2] + 1)
            )), dtype=np.float64
        )
        hkls = hkls[np.argsort(np.linalg.norm(hkls, axis=-1))]
        qs = np.dot(recip, hkls.T).T

        # filter out by theta bounds
        q_mags = np.linalg.norm(qs, axis=-1)
        allowed = np.where(np.logical_and(q_mags <= max_r * 2 * np.pi, q_mags >= min_r * 2 * np.pi))
        qs = qs[allowed]
        hkls = hkls[allowed]

        # compute Bragg condition to find peak locations
        sin_tau = np.linalg.norm(qs, axis=1) * self.wavelength / (4 * np.pi)
        sin_tau[sin_tau > 1] = 0
        taus = 2 * np.arcsin(sin_tau)

        # compute structure factor S(q) as sum of atomic scattering factors
        S_q = np.zeros_like(taus)

        if self.progress:
            import tqdm
            bar = tqdm.tqdm
        else:
            def bar(x):
                return x

        for ind, q_vector in bar(enumerate(qs)):
            # accumulate atomic scattering factors
            atomic_factor = {}
            for species in set(self.doc.atom_types):
                atomic_factor[species] = self.atomic_scattering_factor(q_mags[ind], species)
            factors = np.array([atomic_factor[species] for species in self.doc.atom_types])
            F_s = np.sum(np.exp(1j * positions_abs @ q_vector) * factors * site_occupancies)
            S_q[ind] = np.abs(F_s)**2

        # apply Lorentz correction for polarisation and finite size effects
        S_q *= 2 * (1 + np.cos(taus) ** 2 * np.cos(2 * self.theta_m) ** 2) / (np.sin(taus) * np.sin(0.5 * taus))

        # thermal correction assuming no Debye-Waller factor
        S_q *= np.exp(-np.sin(taus)**2 / self.wavelength**2)**2
        S_q /= np.max(S_q)

        # create histogram and broaden onto 2 theta space in degrees
        self.peak_positions = (180 / np.pi) * taus
        self.hkls = hkls
        self.two_thetas = np.arange(self.two_theta_bounds[0],
                                    self.two_theta_bounds[1] + self.two_theta_resolution,
                                    self.two_theta_resolution)

        self.pattern, bins = np.histogram(self.peak_positions, bins=self.two_thetas, weights=S_q)

        if self.lorentzian_width > 0:
            self.pattern = self._broadening_unrolled(
                self.pattern, self.two_thetas, self.lorentzian_width, broadening_type='lorentzian'
            )
        else:
            # shift and clip the last two theta value if we didnt do broadening
            self.two_thetas = self.two_thetas[:-1] + self.two_theta_resolution / 2

        self.pattern /= np.max(self.pattern)

        order = np.argsort(self.peak_positions)
        self.hkls = hkls[order]
        self.peak_intensities = S_q
        self.peak_positions = self.peak_positions[order]

        # alias old name for compatibility...
        self.spectrum = self.pattern

    def calculate(self):
        """ Alias for calculating the PXRD pattern. """
        self.calc_pxrd()

    def atomic_scattering_factor(self, q_mag, species):
        """ Return fit for particular atom at given q-vector.

        Parameters:
            q_mag (float): magnitude of the q_vector.
            species (str): the element label.

        Returns:
            float: the atomic scattering factor.

        """
        a = self.atomic_scattering_coeffs[species][0]
        b = self.atomic_scattering_coeffs[species][1]
        c = self.atomic_scattering_coeffs[species][2]
        return c + np.sum(a * np.exp(-b * (q_mag / (4 * np.pi))**2))

    def plot(self, **kwargs):
        """ Wrapper function to plot the PXRD pattern. """
        from matador.plotting.pxrd_plotting import plot_pxrd
        plot_pxrd(self, **kwargs)

    def save_pattern(pxrd, fname):
        """ Write a file to `fname` that contains the xy coordinates of
        the PXRD pattern.

        """

        if os.path.isfile(fname):
            raise RuntimeError(f"Requested filename {fname} already exists!")

        from matador import __version__
        header = f""" PXRD pattern computed with matador {__version__}.
Input file:
{pxrd.doc.source[0]}

Structure:
{pxrd.doc}

Settings:
wavelength = {pxrd.wavelength} Å
theta_m = {pxrd.theta_m} degrees
lorentzian_width = {pxrd.lorentzian_width} degrees

2θ (degrees),\t\t\tRelative intensity"""

        np.savetxt(fname, np.vstack([pxrd.two_thetas, pxrd.pattern]).T, header=header, fmt='%.14e', delimiter='\t')

    def save_peaks(pxrd, fname):
        """ Write a file to `fname` that contains the peak list. """

        if os.path.isfile(fname):
            raise RuntimeError(f"Requested filename {fname} already exists!")

        from matador import __version__
        header = f""" PXRD peaks computed with matador {__version__}.
Input file:
{pxrd.doc.source[0]}

Structure:
{pxrd.doc}

Settings:
wavelength = {pxrd.wavelength} Å
theta_m = {pxrd.theta_m} degrees
lorentzian_width = {pxrd.lorentzian_width} degrees

<hkl>,\t\tPeak position (degrees)"""

        np.savetxt(
            fname,
            np.vstack(
                [pxrd.hkls[:, 0], pxrd.hkls[:, 1], pxrd.hkls[:, 2], pxrd.peak_positions]
            ).T,
            header=header,
            fmt=['% d', '% d', '% d', '%.14e'],
            delimiter='\t'
        )


class PXRDFactory(FingerprintFactory):
    fingerprint = PXRD
    default_key = 'pxrd'



# coding: utf-8
# Distributed under the terms of the MIT license.

""" This file implements classes to store and manipulate electronic and
vibrational DOS, with or without projection data.

"""

import warnings
import copy
import functools
import numpy as np
import scipy.integrate
import scipy.interpolate

from matador.orm.orm import DataContainer
from matador.utils.chem_utils import KELVIN_TO_EV, INVERSE_CM_TO_EV

from .dispersion import Dispersion

EPS = 1e-6
MIN_PHONON_FREQ = -0.01
FREQ_CUTOFF = 1e-12


class DensityOfStates(Dispersion, DataContainer):
    """ Generic class for density of states. """

    required_keys = ['dos', 'energies']

    def __init__(self, *args, **kwargs):
        """ Initialise the DOS and trim the DOS data arrays.

        Parameters:
            data (dict/Dispersion): dictionary containing the phonon dos data, or
                a dispersion object to convert.

        """

        if kwargs.get('gaussian_width') is not None:
            self.gaussian_width = kwargs['gaussian_width']

        if args and isinstance(args[0], dict):
            data = args[0]
        else:
            data = kwargs
        # as we can also construct a DOS from arbitarary kpoint/energy data,
        # check that we've been passed this first
        if isinstance(data, Dispersion) or (isinstance(data, dict) and ('dos' not in data and 'spin_dos' not in data)):
            data = self._from_dispersion(data)
        elif isinstance(data, DensityOfStates):
            data = copy.deepcopy(DensityOfStates._data)

        # trigger generation of dos key from spin dos
        if 'dos' not in data and 'spin_dos' in data:
            data["dos"] = np.asarray(data["spin_dos"]["up"]) + np.asarray(data["spin_dos"]["down"])

        super().__init__(data)
        self._trim_dos()

    def _trim_dos(self):
        """ Trim the density of states/frequencies to only include the non-zero
        section of the DOS.

        """
        dos = self._data['dos']
        first_index = np.argmax(dos > EPS)
        last_index = len(dos) - np.argmax(dos[::-1] > EPS)
        self._trimmed_dos = dos[first_index:last_index]
        self._trimmed_energies = self._data['energies'][first_index:last_index]

    def _from_dispersion(self, data, **kwargs):
        """ Convert a Dispersion instance to a DOS. """
        _data = {}
        dos, energies = self.bands_as_dos(data, gaussian_width=self.gaussian_width)

        for key in data:
            _data[key] = data[key]

        _data['dos'] = dos
        _data['energies'] = energies
        return _data

    @property
    def sample_dos(self):
        """ Return the calculated density of states, trimmed at each end to
        only include non-zero values.

        """
        return self._trimmed_dos

    @property
    def sample_energies(self):
        """ Return the energies corresponding to the trimmed DOS. """
        return self._trimmed_energies

    def plot_dos(self, **kwargs):
        """ Plot the density of states. """
        from matador.plotting.spectral_plotting import plot_spectral
        _kwargs = {
            "plot_dos": True,
            "plot_pdos": "pdos" in self,
            "plot_bandstructure": False,
            "phonons": "Vibrational" in self.__class__.__name__
        }
        _kwargs.update(kwargs)
        plot_spectral(
            self,
            **_kwargs
        )

    @staticmethod
    def bands_as_dos(bands, gaussian_width=0.1):
        """ Convert bands data to DOS data. """
        if 'eigs_s_k' in bands:
            eigs_key = 'eigs_s_k'
        elif 'eigs_q' in bands:
            eigs_key = 'eigs_q'
        else:
            raise RuntimeError('Missing eigenvalue keys from bands data.')

        raw_eigs = np.asarray(bands[eigs_key]) - bands.get('fermi_energy', 0)
        raw_weights = np.ones_like(raw_eigs)
        if 'kpoint_weights' in bands:
            for sind, _ in enumerate(bands[eigs_key]):
                for kind, _ in enumerate(bands[eigs_key][sind][0]):
                    raw_weights[sind, :, kind] = bands['kpoint_weights'][kind]

        if len(raw_weights) != 1:
            if len(raw_weights) > 2:
                raise NotImplementedError('Non-collinear spin not supported')
            spin_dos = dict()
            keys = ['up', 'down']
            for sind, _ in enumerate(raw_weights):
                spin_dos[keys[sind]], energies = DensityOfStates._cheap_broaden(
                    bands[eigs_key][sind].flatten(),
                    weights=raw_weights[sind].flatten(),
                    gaussian_width=gaussian_width
                )

                if 'spin_fermi_energy' in bands:
                    energies -= bands['spin_fermi_energy'][0]

            return spin_dos, energies

        dos, energies = DensityOfStates._cheap_broaden(
            raw_eigs.flatten(), weights=raw_weights.flatten(), gaussian_width=gaussian_width
        )

        if 'fermi_energy' in bands:
            energies -= bands['fermi_energy']

        return dos, energies

    @staticmethod
    def _cheap_broaden(eigs, weights=None, gaussian_width=None):
        """ Quickly broaden and bin a set of eigenvalues.

        Parameters:
            eigs (numpy.ndarray): eigenvalue array.
            weights (numpy.ndarray): array of weights.

        Keyword arguments:
            gaussian_width (float): width of gaussian broadening
                to apply.

        Returns:
            Two arrays containing the DOS and energies.

        """
        if gaussian_width is None:
            gaussian_width = 0.1

        hist, energies = np.histogram(eigs, weights=weights, bins=1001)
        if gaussian_width == 0:
            return hist, energies

        # shift bin edges to bin centres
        energies -= energies[1] - energies[0]
        energies = energies[:-1]
        new_energies = np.reshape(energies, (1, len(energies)))
        new_energies = new_energies - np.reshape(energies, (1, len(energies))).T
        dos = np.sum(hist * np.exp(-(new_energies)**2 / gaussian_width), axis=1)
        dos = np.divide(dos, np.sqrt(2 * np.pi * gaussian_width**2))

        return dos, energies


class VibrationalDOS(DensityOfStates):
    """ Specific class for phonon DOS data, including free energy integration. """

    gaussian_width = 10 * INVERSE_CM_TO_EV

    @property
    def debye_temperature(self):
        """ Returns the Debye temperature in K. """
        return self.debye_freq / KELVIN_TO_EV

    @property
    def debye_freq(self):
        """ Returns the Debye frequency in eV. """
        return np.max(self.eigs)

    @property
    def zpe(self):
        """ The zero-point energy per atom as computed from frequency data. """
        if 'zero_point_energy_per_atom' not in self._data:
            if 'eigs_q' not in self._data:
                raise RuntimeError('Unable to compute ZPE without frequency data.')

            zpe = self._compute_zero_point_energy(self.eigs, self.num_kpoints, kpoint_weights=self.kpoint_weights)
            self['zero_point_energy'] = zpe
            self['zero_point_energy_per_atom'] = zpe / (self.num_modes / 3)

        return self._data['zero_point_energy_per_atom']

    @staticmethod
    def _compute_zero_point_energy(eigs, num_kpoints, kpoint_weights=None):
        """ Computes and returns the zero-point energy of the cell
        in eV from frequency data.

        Parameters:
            eigs (np.ndarray): phonon eigenvalues (in eV) array
                in any shape. If `kpoint_weights` is passed, then
                at least one axis of eigs must match the length
                of `kpoint_weights`.
            num_kpoints (int): the number of kpoints at which
                these eigenvalues were calculated.

        Keyword arguments:
            kpoint_weights (np.ndarray): array of weights to use
                for each kpoint.

        """
        min_energy = np.min(eigs)
        if min_energy < MIN_PHONON_FREQ:
            warnings.warn(
                'Imaginary frequency phonons found in this structure {:.1f}, ZPE '
                'calculation will be unreliable, using 0 eV as lower limit of integration.'
                .format(min_energy)
            )

        if kpoint_weights is not None:
            # if kpoint weights are all the same, discard them
            # and just normalise by number of kpoints
            if len(np.unique(kpoint_weights)) == 1:
                kpoint_weights = None
            else:
                eigs_shape = np.shape(eigs)
                if not any(len(kpoint_weights) == axis for axis in eigs_shape):
                    raise RuntimeError(
                        'Unable to match eigs with shape {} with kpoint weights of length {}'
                        .format(eigs_shape, len(kpoint_weights))
                    )
        _eigs = np.copy(eigs)
        if kpoint_weights is not None:
            _eigs = _eigs * kpoint_weights
        else:
            _eigs /= num_kpoints

        return 0.5 * np.sum(np.ma.masked_where(_eigs < 0.0, _eigs, copy=False))

    def vibrational_free_energy(self, temperatures=None):
        """ Computes and returns the vibrational contribution to the free
        energy, including zero-point energy, from the phonon frequencies.

        Parameters:
            temperatures (list): list or array of temperatures to compute
                G(T) at.

        Returns:
            (np.ndarray, np.ndarray): temperature and energy array.

        """

        if temperatures is None:
            temperatures = np.linspace(0, 600, num=5)

        try:
            _ = len(temperatures)
        except TypeError:
            temperatures = [temperatures]

        if 'eigs_q' not in self._data:
            raise RuntimeError('Unable to compute free energies without frequency data.')

        temperatures = np.asarray(temperatures)
        free_energy = np.zeros_like(temperatures, dtype=np.float64)

        min_energy = np.min(self._data['eigs_q'][0])
        if min_energy < MIN_PHONON_FREQ:
            warnings.warn(
                'Imaginary frequency phonons found in this structure {:.1f}, free energy '
                'calculation will be unreliable, using {} eV as lower limit of integration.'
                .format(min_energy, FREQ_CUTOFF)
            )

        for ind, temperature in enumerate(temperatures):
            free_energy[ind] = self.compute_free_energy(temperature)

        if len(temperatures) == 1:
            return free_energy[0]

        return temperatures, free_energy

    @functools.lru_cache(100)
    def compute_free_energy(self, temperature):
        """ Compute the vibrational free energy at the given temperature, using
        lru_cache to avoid doing much extra work. Uses minimum temperature cutoff
        of 1e-9, below which it returns just the ZPE (unless T < 0 K).

        Raises:
            RuntimeError: if temperature is < 0 K.

        Returns:
            float: vibrational free energy per atom, including ZP correction.

        """

        free_energy = 0.0
        kT = KELVIN_TO_EV * temperature
        if temperature < 0.0:
            raise RuntimeError('Not calculating free energies at T = {} K < 0 K'.format(temperature))

        if temperature < 1e-9:
            return self.zpe

        for mode_ind in range(self.num_modes):
            for qpt_ind in range(self.num_qpoints):
                freq = self._data['eigs_q'][0][mode_ind][qpt_ind]
                if freq > FREQ_CUTOFF and freq / kT < 32:
                    contrib = kT * np.log(1 - np.exp(-freq/kT))
                    if 'kpoint_weights' in self._data:
                        contrib *= self.kpoint_weights[qpt_ind]
                    else:
                        contrib /= self.num_qpoints

                    free_energy += contrib

        # normalize by number of atoms
        free_energy /= (self.num_modes / 3)

        # add on zpe per atom
        free_energy += self.zpe

        return free_energy

    def vibrational_free_energy_from_dos(self, temperatures=None):
        """ Computes the vibrational contribution to the free energy
        at a given set of temperatures.

        Keyword arguments:
            temperature (list): list, array or float of temperatures.

        """
        if temperatures is None:
            temperatures = np.linspace(0, 600, num=5)

        temperatures = np.asarray(temperatures)
        free_energy = np.zeros_like(temperatures)
        errs = np.zeros_like(free_energy)

        min_energy = self.sample_energies[0]
        max_energy = self.sample_energies[-1]
        if min_energy < 0.01:
            warnings.warn(
                'Imaginary frequency phonons found in this structure {:.1f}, free energy '
                'calculation will be unreliable, using {} eV as lower limit of integration.'
                .format(min_energy, FREQ_CUTOFF),
                Warning
            )
            min_energy = FREQ_CUTOFF

        for ind, temperature in enumerate(temperatures):

            # if 0 K is requested, return 0 and move on
            if temperature == 0:
                free_energy[ind] = 0.0
                errs[ind] = 0.0
                continue

            kT = KELVIN_TO_EV * temperature

            def integrand(omega):
                return self.vdos_function(omega) * np.log(1 - np.exp(-omega/kT))

            result = scipy.integrate.quad(
                integrand,
                min_energy,
                max_energy
            )

            free_energy[ind] = kT * result[0]
            errs[ind] = result[1]

        if len(temperatures) == 1:
            return free_energy[0]

        return temperatures, free_energy

    @property
    def vdos_function(self):
        """ From the data arrays :attr:`sample_energies` and :attr:`sample_dos`,
        return an interpolated function to integrate.

        """
        return scipy.interpolate.interp1d(
            self.sample_energies,
            self.sample_dos,
            fill_value=(0, 0),
            bounds_error=False,
            copy=False
        )

    def plot_free_energy(self, temperatures=None, ax=None, **kwargs):
        """ Plot G(T) on the array of given temperatures. Default T is [0, 800].

        Keyword arguments:
            temperatures (list/np.ndarray): list or array of temperatures to plot.
                If the array/list has length 2, use these as the start and endpoints
                with 21 plotting points.
            ax (matplotlib.pyplot.Axis): axis object to plot onto.

        """
        from matador.plotting.temperature_plotting import plot_free_energy
        return plot_free_energy(self, temperatures=temperatures, ax=ax, **kwargs)


class ElectronicDOS(DensityOfStates):
    """ Specific class for electronic DOS data. """
    gaussian_width = 0.01



# coding: utf-8
# Distributed under the terms of the MIT License.

""" The workflows module contains ways of creating custom workflows for
chaining together calculations. Each custom workflow inherits from
:obj:`workflows.Workflow` and consists of several
:obj:`workflows.WorkflowStep` objects.
If the creation of the :obj:`Workflow` is wrapped in a function with
the correct signature, a whole :obj:`Workflow` can itself be used as a
:obj:`workflows.WorkflowStep`.


"""


__all__ = ['Workflow', 'WorkflowStep']

__author__ = 'Matthew Evans'
__maintainer__ = 'Matthew Evans'


from matador.workflows.workflows import Workflow, WorkflowStep



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule implements some useful functions for real/reciprocal
cell manipulation, symmetry checking and sampling (e.g. grids and paths.)

"""

import numpy as np
from periodictable import elements
EPS = 1e-12


def abc2cart(lattice_abc):
    """ Converts lattice parameters into Cartesian lattice vectors.

    Parameters:
        lattice_abc (list): [[a, b, c], [alpha, beta, gamma]]

    Returns:
        list: Cartesian lattice vectors.

    """
    assert len(lattice_abc) == 2
    assert len(lattice_abc[0]) == 3
    assert len(lattice_abc[1]) == 3
    a = lattice_abc[0][0]
    b = lattice_abc[0][1]
    c = lattice_abc[0][2]
    deg2rad = np.pi/180
    alpha = lattice_abc[1][0] * deg2rad
    beta = lattice_abc[1][1] * deg2rad
    gamma = lattice_abc[1][2] * deg2rad
    lattice_cart = []
    lattice_cart.append([a, 0.0, 0.0])
    # vec(b) = (b np.cos(gamma), b np.sin(gamma), 0)
    bx = b*np.cos(gamma)
    by = b*np.sin(gamma)
    tol = 1e-12
    if abs(bx) < tol:
        bx = 0.0
    if abs(by) < tol:
        by = 0.0
    cx = c*np.cos(beta)
    if abs(cx) < tol:
        cx = 0.0
    cy = c*(np.cos(alpha)-np.cos(beta)*np.cos(gamma))/np.sin(gamma)
    if abs(cy) < tol:
        cy = 0.0
    cz = np.sqrt(c**2 - cx**2 - cy**2)
    lattice_cart.append([bx, by, 0.0])
    lattice_cart.append([cx, cy, cz])
    return lattice_cart


def cart2abcstar(lattice_cart):
    """ Convert lattice_cart =[[a1,a2,a3],[b1,b2,b3],[c1,c2,c3]]
    to the reciprocal of the lattice vectors, NOT the reciprocal lattice vectors.

    Parameters:
        lattice_cart (list): Cartesian lattice vectors.

    Returns:
        list: lattice parameters [[a,b,c],[alpha,beta,gamma]]

    """
    return np.asarray(real2recip(lattice_cart)) / (2*np.pi)


def cart2volume(lattice_cart):
    """ Convert lattice_cart to cell volume.

    Parameters:
        lattice_cart (list): Cartesian lattice vectors.

    Returns:
        float: cell volume in Angstrom^3.

    """
    lattice_cart = np.asarray(lattice_cart)
    vol = np.abs(np.dot(np.cross(lattice_cart[0], lattice_cart[1]), lattice_cart[2]))
    return vol


def cart2abc(lattice_cart):
    """ Convert Cartesian lattice vectors to lattice parametres.

    Parameters:
        lattice_cart (list): Cartesian lattice vectors.

    Returns:
        list: lattice parameters [[a,b,c],[alpha,beta,gamma]].

    """
    vecs = lattice_cart
    lattice_abc = []
    a, b, c = (np.sqrt(sum([val**2 for val in vec])) for vec in vecs)
    lattice_abc.append([a, b, c])
    # np.cos(alpha) = b.c /|b * c|
    radians2deg = 180.0/np.pi
    cos_alpha = sum([val_b*val_c for (val_c, val_b) in zip(vecs[2], vecs[1])]) / (b*c)
    cos_beta = sum([val_c*val_a for (val_c, val_a) in zip(vecs[2], vecs[0])]) / (c*a)
    cos_gamma = sum([val_a*val_b for (val_a, val_b) in zip(vecs[0], vecs[1])]) / (a*b)
    alpha = radians2deg * np.arccos(cos_alpha)
    beta = radians2deg * np.arccos(cos_beta)
    gamma = radians2deg * np.arccos(cos_gamma)
    lattice_abc.append([alpha, beta, gamma])
    return lattice_abc


def frac2cart(lattice_cart, positions_frac):
    """ Convert positions_frac block into positions_abs.

    Parameters:
        lattice_cart (list): Cartesian lattice vectors.
        positions_frac (list): list of fractional position vectors.

    Returns:
        list: list of absolute position vectors.

    """
    _positions_frac = np.asarray(positions_frac)
    reshaped = False
    if len(np.shape(_positions_frac)) == 1:
        reshaped = True
        _positions_frac = _positions_frac.reshape((1, 3))
    _lattice_cart = np.asarray(lattice_cart)
    positions_abs = switch_coords(_lattice_cart, _positions_frac)
    if reshaped:
        positions_abs = positions_abs.reshape(-1)
    return positions_abs.tolist()


def wrap_frac_coords(positions, remove=False):
    """ Wrap the given fractional coordinates back into the cell.

    Parameters:
        positions (list): list of fractional position vectors, or
            a single position.

    Keyword arguments:
        remove (bool): if True, removes points exterior to the cell.

    Returns:
        list: list of wrapped fractional position vectors.

    """
    from copy import deepcopy
    wrapped = deepcopy(positions)
    single = False
    if len(wrapped) == 3 and isinstance(wrapped[0], float):
        wrapped = [wrapped]
        single = True
    if remove:
        to_remove = len(wrapped)*[False]
    for ind, pos in enumerate(wrapped):
        for k in range(3):
            if pos[k] >= 1 or pos[k] < 0:
                if remove:
                    to_remove[ind] = True
                    break
                else:
                    wrapped[ind][k] %= 1
    if remove:
        wrapped = [wrapped[ind] for ind, res in enumerate(to_remove) if not res]
    if single:
        return wrapped[0]

    return wrapped


def switch_coords(lattice, pos, norm=None):
    """ Act on coordinates with the relevant lattice
    vectors to switch from fractional to absolute coordinates.

    Parameters:

        lattice (np.ndarray(3, 3)): either lattice_cart or reciprocal lattice_cart
        pos (np.ndarray(3, :)): input positions to convert

    Keyword arguments:
        norm (float): divide final coordinates by normalisation factor, e.g.
            2*np.pi when lattice is recip and positions are cartesian.

    Returns:
        np.ndarray(3, :): converted positions

    """
    new_pos = np.zeros_like(pos)
    for ni, _ in enumerate(pos):
        for j in range(3):
            new_pos[ni] += lattice[j] * pos[ni][j]
    if norm is not None:
        new_pos /= norm
    return new_pos


def cart2frac(lattice_cart, positions_abs):
    """ Convert positions_abs block into positions_frac (and equivalent
    in reciprocal space).

    Parameters:
        lattice_cart (list): Cartesian lattice vectors.
        positions_abs (list): list of absolute position vectors.

    Returns:
        list: list of fractional position vectors with the same shape
            as the input list.

    """
    _positions_abs = np.asarray(positions_abs, dtype=np.float64)
    reshaped = False
    if len(np.shape(_positions_abs)) == 1:
        reshaped = True
        _positions_abs = _positions_abs.reshape((1, 3))
    recip_lat = np.asarray(real2recip(lattice_cart))
    recip_lat = recip_lat.T
    positions_frac = switch_coords(recip_lat, _positions_abs, norm=2*np.pi)
    if reshaped:
        positions_frac = positions_frac.reshape(-1)
    return positions_frac.tolist()


def real2recip(real_lat):
    """ Convert the real lattice in Cartesian basis to
    the reciprocal space lattice.

    Parameters:
        real_lat (list): Cartesian lattice vectors.

    Returns:
        list: Cartesian lattice vectors of reciprocal lattice.

    """
    real_lat = np.asarray(real_lat)
    recip_lat = np.zeros((3, 3))
    volume = np.dot(real_lat[0], np.cross(real_lat[1], real_lat[2]))
    recip_lat[0] = (2*np.pi)*np.cross(real_lat[1], real_lat[2])
    recip_lat[1] = (2*np.pi)*np.cross(real_lat[2], real_lat[0])
    recip_lat[2] = (2*np.pi)*np.cross(real_lat[0], real_lat[1])
    recip_lat /= volume
    return recip_lat.tolist()


def calc_mp_grid(lattice_cart, spacing):
    """ Return correct Monkhorst-Pack grid based on lattice
    vectors and desired spacing.

    Parameters:
        lattice_cart (list): Cartesian lattice vectors.
        spacing (float): desired maximum grid spacing.

    Returns:
        list: list of 3 integers defining the MP grid.

    """
    recip_lat = real2recip(lattice_cart)
    recip_len = np.zeros((3))
    recip_len = np.sqrt(np.sum(np.power(recip_lat, 2), axis=1))
    mp_grid = recip_len / (2 * np.pi * spacing)
    return [int(np.ceil(elem)) for elem in mp_grid]


def shift_to_include_gamma(mp_grid):
    """ Calculate the shift required to include $\\Gamma$.
    in the Monkhorst-Pack grid.

    Parameters:
        mp_grid (:obj:`list` of :obj:`int`): number of grid points
            in each reciprocal space direction.

    Returns:
        :obj:`list` of :obj:`float`: shift required to include $\\Gamma$.

    """
    shift = [0, 0, 0]
    for ind, val in enumerate(mp_grid):
        if val % 2 == 0:
            shift[ind] = 1.0/(val*2)
    return shift


def shift_to_exclude_gamma(mp_grid):
    """ Calculate the shift required to exclude $\\Gamma$.
    in the Monkhorst-Pack grid. Returns the "minimal shift", i.e. only
    one direction will be shifted.

    Parameters:
        mp_grid (:obj:`list` of :obj:`int`): number of grid points
            in each reciprocal space direction.

    Returns:
        :obj:`list` of :obj:`float`: shift required to exclude $\\Gamma$.

    """
    shift = [0, 0, 0]
    if all([val % 2 == 1 for val in mp_grid]):
        for ind, val in enumerate(mp_grid):
            if val % 2 == 1:
                shift[ind] = 1.0/(val*2)
                break

    return shift


def get_best_mp_offset_for_cell(doc):
    """ Calculates the "best" kpoint_mp_offset to use for the passed
    cell. If the crystal has a hexagonal space group, then the offset
    returned will shift the grid to include $\\Gamma$ point, and vice
    versa for non-hexagonal cells.

    Parameters:
        doc (dict): matador document to consider, containing structural
            information and a "kpoints_mp_spacing" key.

    Returns:
        :obj:`list` of :obj:`float`: the desired kpoint_mp_offset.

    """

    gamma = False
    if '6' in get_spacegroup_spg(doc, symprec=1e-5):
        gamma = True

    print(doc)

    if 'lattice_cart' not in doc or 'kpoints_mp_spacing' not in doc:
        raise RuntimeError('Unable to calculate offset without lattice or spacing')

    mp_grid = calc_mp_grid(doc['lattice_cart'], doc['kpoints_mp_spacing'])
    if gamma:
        return shift_to_include_gamma(mp_grid)

    return shift_to_exclude_gamma(mp_grid)


def calc_mp_spacing(real_lat, mp_grid, prec=3):
    """ Convert real lattice in Cartesian basis and the
    kpoint_mp_grid into a grid spacing.

    Parameters:
        real_lat (list): Cartesian lattice vectors.
        mp_grid (:obj:`list` of :obj:`int`): 3 integers defining the MP grid.

    Keyword arguments:
        prec (int): desired decimal precision of output.

    Returns:
        float: mp_spacing rounded to `prec`.

    """
    recip_lat = real2recip(real_lat)
    recip_len = np.zeros((3))
    recip_len = np.sqrt(np.sum(np.power(recip_lat, 2), axis=1))
    spacing = recip_len / (2*np.pi*np.asarray(mp_grid))
    max_spacing = np.max(spacing)
    exponent = round(np.log10(max_spacing) - prec)
    return round(max_spacing + 0.5*10**exponent, prec)


def get_seekpath_kpoint_path(
    doc, standardize=True, explicit=True, spacing=0.01, threshold=1e-7, debug=False, symmetry_tol=None
):
    """ Return the conventional kpoint path of the relevant crystal system
    according to the definitions by "HKPOT" in
    Comp. Mat. Sci. 128, 2017:

    http://dx.doi.org/10.1016/j.commatsci.2016.10.015

    Parameters:
        doc (dict/tuple): matador doc or spglib tuple to find kpoint path for.

    Keyword arguments:
        spacing (float): desired kpoint spacing
        threshold (float): internal seekpath threshold
        symmetry_tol (float): spglib symmetry tolerance

    Returns:
        dict: standardized version of input doc
        list: list of kpoint positions
        dict: full dictionary of all seekpath results

    """
    try:
        from seekpath import get_explicit_k_path, get_path
    except ImportError:
        raise ImportError("SeeK-Path dependency missing, please install it with `pip install seekpath`.")

    if symmetry_tol is None:
        symmetry_tol = 1e-5

    if isinstance(doc, tuple):
        spg_structure = doc
    else:
        if standardize:
            spg_structure = doc2spg(standardize_doc_cell(doc, symprec=symmetry_tol))
        else:
            spg_structure = doc2spg(doc)

    if explicit:
        seekpath_results = get_explicit_k_path(spg_structure,
                                               reference_distance=spacing,
                                               with_time_reversal=True,
                                               symprec=symmetry_tol,
                                               threshold=threshold)

        kpt_path = seekpath_results['explicit_kpoints_rel']
    else:
        seekpath_results = get_path(spg_structure)
        kpt_path = []

    primitive_doc = dict()
    primitive_doc['lattice_cart'] = seekpath_results['primitive_lattice']
    primitive_doc['positions_frac'] = seekpath_results['primitive_positions']
    primitive_doc['atom_types'] = [str(elements[i]) for i in seekpath_results['primitive_types']]
    primitive_doc['num_atoms'] = len(primitive_doc['atom_types'])
    primitive_doc['lattice_abc'] = cart2abc(primitive_doc['lattice_cart'])
    primitive_doc['cell_volume'] = cart2volume(primitive_doc['lattice_cart'])
    if debug:
        print('Found lattice type {}'.format(seekpath_results['bravais_lattice_extended']))
        print('Old lattice:\n', np.asarray(doc['lattice_cart']))
        print('Contained {} atoms'.format(doc['num_atoms']))
        print('New lattice:\n', np.asarray(primitive_doc['lattice_cart']))
        print('Contains {} atoms'.format(primitive_doc['num_atoms']))
        print('k-point path contains {} points.'.format(len(kpt_path)))

    if 'site_occupancy' in doc:
        if min(doc['site_occupancy']) < 1 - EPS:
            print('Ignoring any site occupancy found in this cell.')
        primitive_doc['site_occupancy'] = [1 for atom in primitive_doc['atom_types']]

    return primitive_doc, kpt_path, seekpath_results


def doc2spg(doc, check_occ=True):
    """ Return an spglib input tuple from a matador doc.

    Parameters:
        doc (dict or :class:`Crystal`): matador document or Crystal object.

    Keyword arguments:
        check_occ (bool): check for partial occupancy and raise an error if
            present.

    Returns:
        tuple: spglib-style tuple of lattice, positions and types.

    """
    from matador.utils.chem_utils import get_atomic_number
    try:
        if 'lattice_cart' not in doc:
            doc['lattice_cart'] = abc2cart(doc['lattice_abc'])
    except KeyError:
        raise RuntimeError("doc2spg failed: unable to find lattice in document.")

    required_keys = ("lattice_cart", "positions_frac", "atom_types")
    if all(key in doc for key in required_keys):
        if 'lattice_cart' not in doc:
            doc['lattice_cart'] = abc2cart(doc['lattice_abc'])
        cell = (doc['lattice_cart'],
                doc['positions_frac'],
                [get_atomic_number(elem) for elem in doc['atom_types']])
        if check_occ and np.min(doc.get('site_occupancy', [1.0])) < 1.0:
            raise RuntimeError("spglib does not support partial occupancy.")

        return cell

    else:
        raise RuntimeError(f"Unable to use doc2spg, one of {required_keys} was missing.")


def get_space_group_label_latex(label):
    """ Return the LaTeX format of the passed space group label. Takes
    any string, leaves the first character upright, italicses the rest,
    handles subscripts and bars over numbers.

    Parameters:
        label (str): a given space group in "standard" plain text format,
        e.g. P-63m.

    Returns:
        str: the best attempt to convert the label to LaTeX.

    """
    latex_label = '$'
    if not isinstance(label, str):
        raise RuntimeError("Space group label must be a string, not {}".format(label))

    skip = False
    for ind, char in enumerate(label):
        if skip:
            skip = False
            continue

        if char == '-':
            # add the next char inside the bar
            latex_label += "\\bar{" + label[ind+1] + "}"
            skip = True

        else:
            latex_label += char

    latex_label += "$"

    return latex_label


def standardize_doc_cell(doc, primitive=True, symprec=1e-2):
    """ Return standardized cell data from matador doc.

    Parameters:
        doc (dict or :class:`Crystal`): matador document or Crystal object.

    Keyword arguments:
        primitive (bool): whether to reduce cell to primitive.
        symprec (float): spglib symmetry tolerance.

    Returns:
        dict: matador document containing standardized cell.

    """
    import spglib as spg
    from matador.crystal import Crystal
    from matador.utils.chem_utils import get_atomic_symbol
    from copy import deepcopy

    spg_cell = doc2spg(doc)
    spg_standardized = spg.standardize_cell(spg_cell, to_primitive=primitive, symprec=symprec)
    if not isinstance(doc, Crystal):
        std_doc = deepcopy(doc)
    else:
        std_doc = deepcopy(doc._data)
    std_doc['lattice_cart'] = [list(vec) for vec in spg_standardized[0]]
    std_doc['lattice_abc'] = cart2abc(std_doc['lattice_cart'])
    std_doc['positions_frac'] = [list(atom) for atom in spg_standardized[1]]
    std_doc['atom_types'] = [get_atomic_symbol(atom) for atom in spg_standardized[2]]
    std_doc['site_occupancy'] = len(std_doc['positions_frac']) * [1]
    std_doc['cell_volume'] = cart2volume(std_doc['lattice_cart'])
    std_doc['space_group'] = get_spacegroup_spg(std_doc, symprec=symprec)
    # if the original document was a crystal, return a new one
    if isinstance(doc, Crystal):
        std_doc = Crystal(std_doc)

    return std_doc


def get_spacegroup_spg(doc, symprec=0.01, check_occ=True):
    """ Return spglib spacegroup for a cell.

    Parameters:
        doc (dict or :class:`Crystal`): matador document or Crystal object.

    Keyword arguments:
        symprec (float): spglib symmetry tolerance.

    Returns:
        str: spacegroup symbol of structure.

    """
    import spglib as spg
    spg_cell = doc2spg(doc, check_occ=check_occ)
    space_group = spg.get_spacegroup(spg_cell, symprec=symprec)
    if space_group is None:
        raise RuntimeError('Spglib was unable to calculate space group.')

    return space_group.split(' ')[0]


def add_noise(doc, amplitude=0.1):
    """ Add random noise to the positions of structure contained in doc.
    Useful for force convergence tests.

    Parameters:
        doc (dict): dictionary containing matador structure.

    Keyword arguments:
        amplitude (float): maximum amplitude of noise vector.

    Raises:
        KeyError if (`lattice_cart` and `positions_frac`) or `positions_abs`
            are missing.

    Returns:
        dict: the randomised structure.
    """
    poscart = np.asarray(doc.get('positions_abs') or frac2cart(doc['lattice_cart'], doc['positions_frac']))
    for atom in poscart:
        noise_vector = np.random.rand(3)
        noise_vector /= np.sqrt(np.sum(noise_vector**2))
        noise_vector *= amplitude*(np.random.rand())
        atom += noise_vector
    doc['positions_abs'] = poscart.tolist()
    doc['positions_frac'] = cart2frac(doc['lattice_cart'], doc['positions_abs'])

    return doc


def calc_pairwise_distances_pbc(poscart, images, lattice, rmax,
                                poscart_b=None, compress=False, debug=False,
                                filter_zero=False, per_image=False):
    """ Calculate PBC distances with SciPy's cdist, given the
    image cell vectors.

    Parameters:
        poscart (numpy.ndarray): list or array of absolute atomic coordinates.
        images: iterable of lattice vector multiples (e.g. [2, -1, 3])
            required to obtain the translation to desired image cells.
        lattice (:obj:`list` if :obj:`list`): list of lattice vectors of
            the real cell.
        rmax (float): maximum value after which to mask the array.

    Keyword arguments:
        poscart_b (numpy.ndarray): absolute positions of another type of
            atom, where only A-B distances will be calculated.
        debug (bool): print timing data and how many distances were masked.
        compress (bool): whether or not to compressed the output array,
            useful when e.g. creating PDFs but not when atom ID is important.
        filter_zero (bool): whether or not to filter out the "self-interaction"
            zero distances.
        per_image (bool): return a list of distances per image, as opposed to
            one large flat. This preserves atom IDs for use elsewhere.

    Returns:
        distances (numpy.ndarray): pairwise 2-D d_ij masked array with values
            or stripped 1-D array containing just the distances, or a list of
            numpy arrays if per_image is True.

    """
    from scipy.spatial.distance import cdist
    import time
    if debug:
        start = time.time()
    _lattice = np.asarray(lattice)
    _poscart = np.asarray(poscart)
    image_distances = []
    if poscart_b is None:
        _poscart_b = _poscart
    else:
        _poscart_b = np.asarray(poscart_b)

    num_pairs = len(_poscart) * len(_poscart_b)

    if per_image:
        for image_ind, prod in enumerate(images):
            distances = cdist(_poscart, _poscart_b + prod @ _lattice)
            distances = np.ma.masked_where(distances > rmax, distances, copy=False)
            image_distances.append(distances)

        distances = image_distances

    else:
        array_size = (len(_poscart) * len(images) * len(_poscart_b))
        distances = np.empty(array_size)
        for image_ind, prod in enumerate(images):
            distances[image_ind*num_pairs:(image_ind+1)*num_pairs] = cdist(_poscart, _poscart_b + prod @ _lattice).flatten()

        distances = np.ma.masked_where(distances > rmax, distances, copy=False)
        if filter_zero:
            distances = np.ma.masked_where(distances < EPS, distances, copy=False)

        if debug:
            print('Calculated: {}, Used: {}, Ignored: {}'.format(len(distances),
                                                                 np.ma.count(distances),
                                                                 np.ma.count_masked(distances)))
        if compress:
            distances = distances.compressed()

    if debug:
        end = time.time()
        print('Calculated distances in {} s'.format(end - start))

    return distances


def create_simple_supercell(doc, extension, standardize=False, symmetric=False):
    """ Return a document with new supercell, given extension vector.

    Parameters:
        doc (dict): matador doc to construct cell from.
        extension (:obj:`tuple` of :obj:`int`): multiplicity of each lattice vector,
            e.g. (2,2,1).

    Keyword arguments:
        standardize (bool): whether or not to use spglib to standardize the cell first.
        symmetric (bool): whether or not centre the new cell on the origin.

    Returns:
        supercell_doc (dict): matador document containing supercell.

    """
    from itertools import product
    from copy import deepcopy
    if not all([elem >= 1 for elem in extension]) or extension == (1, 1, 1):
        raise RuntimeError('Bad/null supercell {} requested...'.format(extension))

    supercell_doc = deepcopy(doc)
    # standardize cell with spglib
    if standardize:
        supercell_doc = standardize_doc_cell(supercell_doc)

    # copy new doc and delete data that will not be corrected
    keys_to_copy = set(['positions_frac', 'lattice_cart', 'atom_types', 'source', 'stoichiometry'])
    supercell_doc = {key: supercell_doc[key] for key in keys_to_copy}

    for i, elem in enumerate(extension):
        for k in range(3):
            supercell_doc['lattice_cart'][i][k] = elem * supercell_doc['lattice_cart'][i][k]
        for ind, atom in enumerate(supercell_doc['positions_frac']):
            supercell_doc['positions_frac'][ind][i] /= elem

    images = product(*[list(range(elem)) for elem in extension])
    new_positions = []
    new_atoms = []
    for image in images:
        if image != (0, 0, 0):
            for ind, atom in enumerate(supercell_doc['atom_types']):
                new_pos = deepcopy(supercell_doc['positions_frac'][ind])
                for i, elem in enumerate(image):
                    new_pos[i] += image[i] / extension[i]
                new_positions.append(new_pos)
                new_atoms.append(atom)
    supercell_doc['atom_types'].extend(new_atoms)
    supercell_doc['positions_frac'].extend(new_positions)

    if symmetric:
        supercell_doc['positions_frac'] = np.asarray(supercell_doc['positions_frac'])
        supercell_doc['positions_frac'] -= np.mean(supercell_doc['positions_frac'], axis=0)
        supercell_doc['positions_frac'] = supercell_doc['positions_frac'].tolist()

    supercell_doc['num_atoms'] = len(supercell_doc['atom_types'])
    supercell_doc['cell_volume'] = cart2volume(supercell_doc['lattice_cart'])
    supercell_doc['lattice_abc'] = cart2abc(supercell_doc['lattice_cart'])

    assert np.isclose(supercell_doc['cell_volume'],
                      np.prod(extension)*cart2volume(doc['lattice_cart']))

    return supercell_doc
